import json
import os

import cv2
import matplotlib.pyplot as plt
import vltk.vars as vltk
from tqdm import tqdm
from vltk import adapters
from vltk.features import Features


class FUNSD(adapters.VisnDataset):

    urls = "https://guillaumejaume.github.io/FUNSD/dataset.zip"

    @staticmethod
    def schema():
        return {
            vltk.tokenbox: Features.Box(),
            vltk.text: Features.StringList(),
            vltk.label: Features.StringList(),
            # "linking": Features.NestedIntList,
        }

    @staticmethod
    def forward(json_files, splits, datadir=None):
        imgids = set()
        annos = []
        for filename, data in tqdm(json_files.items()):
            text = []
            words = []
            labels = []
            boxes = []
            linkings = []
            imgid = filename.split(".")[0]

            assert imgid not in imgids
            imgids.update([imgid])

            for item in data["form"]:
                label = item["label"]
                if label not in ("question", "answer", "other"):
                    label = "other"
                linking = item["linking"]
                if not linking:
                    linking = [[0, 0]]
                words = item["words"]
                labels += [label] * len(words)
                linkings += linking * len(words)

                for word in words:
                    text.append(word["text"])
                    x1, y1, x2, y2 = word["box"]
                    boxes.append([x1, y1, x2 - x1, y2 - y1])

            assert len(labels) == len(text)

            entry = {
                vltk.text: text,
                vltk.tokenbox: boxes,
                vltk.label: labels,
                vltk.imgid: str(imgid),
            }
            annos.append(entry)

        return annos



import collections
import contextlib
import importlib
import inspect
import json
import os
import smtplib
import subprocess
import sys
from collections import defaultdict
from collections.abc import Iterable
from datetime import datetime
from email.message import EmailMessage
from typing import Tuple, Union

import datasets
import jsonlines
import numpy as np
import pyarrow
import torch
import yaml
from torch import nn

PATH = os.path.join(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..")), "libdata"
)


# merge dictionaries
def mergedicts(dict1, dict2):
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(mergedicts(dict1[k], dict2[k])))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert you of value conflicts
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


def update_config_with_logdir(config, flags, name, datasets):
    if "base_logdir" not in flags:
        baselogdir = config.base_logdir
    else:
        baselogdir = flags.pop("base_logdir")
    if "rel_logdir" not in flags:
        specifications = ["year", "month", "day", "hour"]
        date = datetime.now()
        date = ":".join([str(getattr(date, s)) for s in specifications])
        rellogdir = f"{name}_" + date
    else:
        rellogdir = flags.pop("rel_logdir")
    if rellogdir == baselogdir:
        rellogdir = ""

    config.update(
        {
            "logdir": os.path.join(baselogdir, rellogdir),
            "rel_logdir": rellogdir,
            "base_logdir": baselogdir,
        }
    )


# this is to update metadata on dataset objects
def set_metadata(tbl, tbl_meta={}):
    fields = []
    for i, f in enumerate(tbl.schema.names):
        fields.append(tbl.schema[i])

    tbl_metadata = tbl.schema.metadata
    for k, v in tbl_meta.items():
        if isinstance(v, dict):
            tbl_metadata[k] = json.dumps(v).encode("utf-8")
        elif isinstance(v, set):
            tbl_metadata[k] = "\n".join(v).encode("utf-8")
        else:
            tbl_metadata[k] = str(v).encode("utf-8")

    schema = pyarrow.schema(fields, metadata=tbl_metadata)
    tbl = pyarrow.Table.from_arrays(list(tbl.itercolumns()), schema=schema)

    return tbl


def batcher(iterable, n=64):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def try_load(filepath):
    ext = str(filepath).split(".")[-1]
    if ext in ("json", "jsonl"):
        try:
            with open(filepath) as f:
                data = json.load(f)
            return data
        except json.decoder.JSONDecodeError:
            with open(filepath) as f:
                data = jsonlines.open(f)
            return data
    elif "pdf" == ext:
        return str(filepath)
    raise Exception(ext, filepath)
    # try:
    #     with open(filepath) as f:
    #         entry = jsonlines.open(f)
    #         yield entry
    # except Exception:
    #     with open(filepath) as f:
    #         data = json.load(f)
    #     return data


def clean_imgid(img_id):
    return str(img_id).replace(" ", "")


def load_arrow(dset_to_arrow_fp: dict, fields: Union[Tuple[str], str, None] = None):
    if fields is not None and not fields:
        return None
    arrow_dict = {}
    for dset in dset_to_arrow_fp:
        arrow_fp = dset_to_arrow_fp[dset]
        arrow = datasets.Dataset.from_file(arrow_fp)
        if fields is not None:
            fields = list(fields)
        arrow.set_format(type="numpy", columns=fields)
        arrow_dict[dset] = arrow
    return arrow_dict


def clip_img_ids(img_ids, percent_data=1.0):
    if percent_data != 1.0:
        stop_int = max(1, int(np.ceil(len(img_ids) * percent_data)))
        img_ids = img_ids[:stop_int]
    assert len(img_ids) > 0
    return img_ids


@contextlib.contextmanager
def dummy_context():
    yield None


def send_email(address, message, failure=True):
    sender = os.environ.get("HOSTNAME", "localhost")
    msg = EmailMessage()
    msg.set_content(message)
    if failure:
        msg["Subject"] = "MLLIB failure!"
    else:
        msg["Subject"] = "MLLIB success!"
    msg["From"] = sender
    msg["To"] = [address]
    s = smtplib.SMTP("localhost")
    s.send_message(msg)
    s.quit()


def unflatten_dict(dictionary):
    resultDict = dict()
    for key, value in dictionary.items():
        parts = key.split(".")
        d = resultDict
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            d = d[part]
        d[parts[-1]] = value
    return resultDict


def load_yaml(flags: dict):
    # allow flags to overwrite keys present in yaml file
    if "yaml" in flags:
        yaml_path = flags.pop("yaml")
        loaded = yaml.load(open(yaml_path), Loader=yaml.Loader)
        updated_flags = dict(mergedicts(loaded, flags))
        return updated_flags
    else:
        return flags


def convert_jax_to_torch_weights(
    torch_dict, jax_dict, torch_encoder_weight_name="transformer"
):

    torch_to_jax_alias = [
        ("bias", "bias"),
        ("norm1", "LayerNorm_0"),
        ("norm2", "LayerNorm_2"),
        ("norm", "encoder_norm"),
        ("weight", "kernel"),
        ("weight", "scale"),
        ("transformer", "Transformer"),
        ("encoder_layers.", "encoderblock_"),
        ("attn", "MultiHeadDotProductAttention_1"),
        ("out", "out"),
        ("query", "query"),
        ("key", "key"),
        ("value", "value"),
        ("mlp", "MlpBlock_3"),
        ("fc1", "Dense_0"),
        ("fc2", "Dense_1"),
        ("pos_embedding", "posembed_input"),
        ("cls_token", "cls"),
        ("classifier", "head"),
    ]
    # pos embdedding = n_patches +  1 (for cls token)

    jax_flattened_weights = flatten_dict(jax_dict, parent_key="")
    jax_dict_renamed = collections.OrderedDict()
    for k, v in jax_flattened_weights.items():
        for (tn, jn) in torch_to_jax_alias:
            k = k.replace(jn, tn)
        jax_dict_renamed[k] = torch.tensor(v.tolist())  # .tolist()
    for j, t in zip(sorted(jax_dict_renamed), sorted(torch_dict)):
        if j != t:
            print(j, t)
        if jax_dict_renamed[j].shape != torch_dict[t].shape:
            jshape = list(jax_dict_renamed[j].shape)
            tshape = list(torch_dict[t].shape)
            assert len(jshape) == len(tshape)
            if sum(jshape) == sum(tshape):
                ind_map = [0] * len(jshape)
                seen_inds = set()
                added_inds = set()
                for ji, jv in enumerate(jshape):
                    for ti, tv in enumerate(tshape):
                        if jv == tv:
                            if ji not in seen_inds and ti not in added_inds:
                                ind_map[ti] = ji
                                seen_inds.add(ji)
                                added_inds.add(ti)
                try:
                    new_val = jax_dict_renamed[j].permute(*tuple(ind_map))
                except Exception:
                    raise Exception(
                        ind_map, jax_dict_renamed[j].shape, torch_dict[j].shape, j
                    )
                assert new_val.shape == torch_dict[t].shape, (
                    new_val.shape,
                    torch_dict[t].shape,
                    ind_map,
                    jshape,
                )
                jax_dict_renamed[j] = new_val
            else:
                print(f"SKIPPIG: mismatched {j, t}, shapes {jshape, tshape}")
                jax_dict_renamed[j] = torch_dict[t]
    assert len([x for x in jax_dict_renamed if torch_encoder_weight_name in x]) == len(
        [x for x in torch_dict if torch_encoder_weight_name in x]
    )
    return jax_dict_renamed


def flatten_dict(d, parent_key="", sep="."):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return collections.OrderedDict(items)


# tensor equality
def tensor_equality(a, b):

    n1 = a.numpy()
    n2 = b.numpy()
    print(n1.shape)
    print(n2.shape)
    print(n1[0, 0, :5])
    print(n2[0, 0, :5])
    assert np.allclose(
        n1, n2, rtol=0.01, atol=0.1
    ), f"{sum([1 for x in np.isclose(n1, n2, rtol=0.01, atol=0.1).flatten() if x == False])/len(n1.flatten())*100:.4f} % element-wise mismatch"
    raise Exception("tensors are all good")


def isprimitive(obj):
    if (
        isinstance(obj, int)
        or isinstance(obj, bool)
        or isinstance(obj, str)
        or isinstance(obj, float)
    ):
        return True
    else:
        return False


def on_children(obj: object, findtype=int, func=None):
    # if object is dict
    if isinstance(obj, dict) and not isinstance(obj, findtype):
        for k, v in obj.items():
            new_v = on_children(v, findtype=findtype, func=func)
            if new_v is not None:
                obj[k] = new_v
    # if the object is a primitive
    elif isprimitive(obj) and not isinstance(obj, findtype):
        return None
    # if the object is iterable
    elif isinstance(obj, Iterable) and not isinstance(obj, findtype):
        for idx, v in enumerate(obj):
            new_v = on_children(v, findtype=findtype, func=func)
            if new_v is not None:
                obj[idx] = new_v
    # if the object is not a findtype, iterable, dictionary, or prim
    elif not isinstance(obj, findtype) and hasattr(obj, "__dict__"):
        raise Exception
    # if the object is the find type
    elif func is not None:
        out = func(obj)
        if out is not None:
            obj = out
        else:
            obj = None
    return obj


def change_device(batch, device="cpu"):
    assert isinstance(device, int) or device == "cpu"
    device = torch.device(device)
    return on_children(batch, findtype=torch.Tensor, func=lambda x: x.to(device))


def check_device(batch):
    return on_children(
        batch,
        findtype=torch.Tensor,
        func=lambda x: (print(x.shape, x.device) if x is not None else x),
    )


def get_list_primitive(ls):

    if isinstance(ls, collections.Iterable) and not isinstance(ls, str):
        return get_list_primitive(ls[0])
    else:
        if ls is None:
            return None
        else:
            return type(ls)


# stack overflow credit
def flatten_stringlist(container):
    if container is None:
        return []
    if isinstance(container, str):
        return [container]
    for i in container:
        if isinstance(i, (list, tuple)):
            for j in flatten_stringlist(i):
                yield j
        else:
            yield i


def get_arrow_primitive(schema_value):
    if hasattr(schema_value, "feature"):
        return get_arrow_primitive(schema_value.feature)
    else:
        return schema_value.dtype


def convertids_recursive(ls, objids):
    ls = list(ls)
    # get deepest nested list
    if (
        isinstance(ls, collections.Iterable)
        and not isinstance(ls, str)
        and isinstance(ls[0], str)
    ):
        ten = torch.Tensor(list(map(lambda x: objids[x], ls)))
        return ten
    else:
        for idx, item in enumerate(ls):
            res = convertids_recursive(item, objids)
            assert isinstance(res, torch.Tensor), (type(res), res)
            ls[idx] = res
        try:
            ls = torch.stack(ls)
        except Exception:
            pass
        return ls



import torch
from vltk.inspection import collect_args_to_func


class Processor:
    _type = None
    _keys = ()

    @property
    def keys(self):
        if isinstance(self._keys, str):
            return set([self._keys])
        return set(self._keys)

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @torch.no_grad()
    def __call__(self, inp, **kwargs):
        if isinstance(inp, dict):
            proc_keys = self.keys
            intersection = proc_keys.intersection(set(inp.keys()))
            assert (
                intersection == proc_keys
            ), f"{type(self).__name__} requires {proc_keys} to be present within the input dictionary, but not all \
                    keys are present. the input dictionary only has: {inp.keys()}"

        kwargs = collect_args_to_func(self.forward, kwargs)
        output = self.forward(inp, **kwargs)
        if not isinstance(output, dict):
            assert isinstance(
                output, torch.Tensor
            ), "the outputs of any processor must be a torch tensor or a \
            dictionary where the repective value(s) from the key(s) of interest, specified in the init method, \
            must be a torch tensor aswell"
        else:
            pass

        return output


class VisnProcessor(Processor):
    _type = "visn"


class LangProcessor(Processor):
    _type = "lang"


class VisnLangProcessor(Processor):
    _type = "visnlang"

    @torch.no_grad()
    def __call__(self, text_inp, visn_inp, **kwargs):

        text_inp, visn_inp = self.forward(text_inp, visn_inp, **kwargs)

        return text_inp, visn_inp



import json
import os
from collections import defaultdict

import numpy as np
import vltk.vars as vltk
from tqdm import tqdm
from vltk.features import Features
from vltk import adapters
from vltk.utils.adapters import get_span_via_jaccard

"""
OCR Results

'status': str,
'recognitionResults': list (each item is a page)
    'page': int, -> keep
    'clockwiseOrientation': float, -> discard
    'width': int, -> discard
    'height': int,-> discard
    'unit': string, -> discard
        'lines': list (each item is a component)
        'boundingBox': list of ints,
        'text': string,
        'words':
            'boundingBox': list of ints,
                        'text': str,
                                'confidence': str  (optional)
"""


class DocVQA(adapters.VisnLangDataset):
    data_info = {
        "val": {"docvqavisn": ["val"]},
        "train": {"docvqavisn": ["train"]},
    }

    @staticmethod
    def format_box(box):
        x1, y1, x2, y2, x3, y3, x4, y4 = box
        new_x1 = min([x1, x2, x3, x4])
        new_x2 = max([x1, x2, x3, x4])
        new_y1 = min([y1, y2, y3, y4])
        new_y2 = max([y1, y2, y3, y4])
        width = abs(new_x2 - new_x1)
        height = abs(new_y2 - new_y1)
        return [x1, y1, width, height]

    @staticmethod
    def schema():
        # img id, label, and score are assumed to be default features
        return {vltk.label: Features.StringList()}  # vltk.span: Features.Span,

    @staticmethod
    def forward(json_files, split, datadir=None):
        total = 0
        skipped = 0
        batch_entries = []
        dataa = defaultdict(list)
        for filename, item in json_files.items():
            data = item["data"]
            for d in tqdm(data):
                split = d["data_split"]
                question = d["question"].lower().replace('"', "")
                cont = True
                for x in (
                    "location",
                    "address",
                    "site",
                    "city",
                    "state",
                    "zip code",
                    "street",
                ):
                    if (
                        x in question
                        and "email" not in question
                        and "value" not in question
                        and "how much" not in question
                        and "percentage" not in question
                        and "to who" not in question
                        and "does this" not in question
                        and "what type" not in question
                        and "whom" not in question
                        and "when" not in question
                        and "last, first" not in question
                        and "date" not in question
                        and "name of a" not in question
                        and "name of c" not in question
                        and "name of m" not in question
                        and "addressed" not in question
                        and "is the drug" not in question
                        and "modifications" not in question
                        and "who" not in question
                        and "website" not in question
                        and len(question.split()) < 12
                    ):
                        cont = False
                        print(question)
                        total += 1
                        break
                if cont:
                    continue
                image = d["image"]
                # docid = d["docId"]
                imgid = image.split(".")[0].split("/")[-1]
                # open annotation:
                fileid = imgid + ".png"

                answers = list(map(lambda x: x.lower(), d["answers"]))
                anno = json.load(
                    open(
                        os.path.join(
                            datadir,
                            "docvqavisn",
                            split,
                            "ocr_results",
                            f"{imgid}.json",
                        ),
                        "r",
                    )
                )["recognitionResults"][0]

                words = ()
                answers = [answers[0]]
                for lines in anno["lines"]:
                    for word in lines["words"]:
                        words += (word["text"].lower(),)

                # if not words:
                #     skipped += 1
                #     continue

                span, max_jaccard = get_span_via_jaccard(words, answers, skipped)
                if span is not None:
                    dataa[fileid].append(span)
                # if span is None:
                #     continue

                entry = {
                    vltk.text: question,
                    vltk.imgid: imgid,
                    vltk.label: answers
                    # vltk.span: span
                }
                batch_entries.append(entry)
        print(f"skipped {skipped} questions: could not find answer.")
        json.dump(dataa, open(f"docvqa_{split}.json", "w"))
        raise Exception(total)
        return batch_entries


class DocVQAVisn(adapters.VisnDataset):
    @staticmethod
    def schema():
        return {
            vltk.box: Features.Box(),
            vltk.tokenbox: Features.Box(),
            vltk.text: Features.StringList(),
        }

    @staticmethod
    def format_box(box):
        x1, y1, x2, y2, x3, y3, x4, y4 = box
        new_x1 = min([x1, x2, x3, x4])
        new_x2 = max([x1, x2, x3, x4])
        new_y1 = min([y1, y2, y3, y4])
        new_y2 = max([y1, y2, y3, y4])
        width = abs(new_x2 - new_x1)
        height = abs(new_y2 - new_y1)
        return [x1, y1, width, height]

    @staticmethod
    def forward(json_files, splits, datadir=None):
        imgids = set()
        annos = []
        for filename, data in tqdm(json_files.items()):
            entry = {}
            imgid = filename.split(".")[0].split("/")[-1]
            assert imgid not in imgids
            imgids.add(imgid)
            status = 1 if data["status"] == "Succeeded" else 0
            if status == 0:
                continue
            data = data["recognitionResults"]
            if len(data) != 1:
                raise Exception(len(data))
            data = data[0]
            boxes = []
            tokenboxes = []
            texts = []
            for lines in data["lines"]:
                box = DocVQAVisn.format_box(lines["boundingBox"])
                boxes.append(box)
                for word in lines["words"]:
                    text = word["text"]
                    box = word["boundingBox"]
                    box = DocVQAVisn.format_box(lines["boundingBox"])
                    texts.append(text)
                    tokenboxes.append(box)
            entry = {
                vltk.imgid: imgid,
                vltk.box: boxes,
                vltk.text: texts,
                vltk.tokenbox: tokenboxes,
            }
            annos.append(entry)

        return annos



import os

from transformers import RobertaTokenizerFast
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig, VisionConfig
from vltk import build

if __name__ == "__main__":
    datadir = os.path.join(os.environ["HOME"], "demodata")
    Adapters().get("docvqa").extract(datadir)
    # Adapters().get("docvqavisn").extract(datadir)
    # config = DataConfig(
    #     # lang=LangConfig(tokenizer="BertWordPieceTokenizer"),
    #     lang=LangConfig(
    #         tokenizer=RobertaTokenizerFast, vocab_path_or_name="roberta-base"
    #     ),
    #     visn=VisionConfig(grayscale=True),
    #     train_datasets=[["docvqa", "trainval"]],
    #     visn_processors=["auxtokenize", "ocrbox"],
    #     visnlang_processors=["span"],
    #     extractor=None,
    #     datadir=datadir,
    #     train_batch_size=2,
    #     eval_batch_size=2,
    #     img_first=True,
    #     num_workers=1,
    # )

    # train_loader, val_loader = build(config)
    # for x in train_loader:
    #     print(x)
    #     break



import datasets as ds


class Features:
    @staticmethod
    def Boxes():
        return ds.Sequence(
            length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("float32"))
        )

    # legacy
    @staticmethod
    def Box():
        return ds.Sequence(
            length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("float32"))
        )

    @staticmethod
    def BoolList():
        return ds.Sequence(length=-1, feature=ds.Value("bool"))

    @staticmethod
    def Polygons():
        return ds.Sequence(
            length=-1,
            feature=ds.Sequence(
                length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("float32"))
            ),
        )

    # def RLE():
    #   ds.Sequence(length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("float32")))

    @staticmethod
    def RLE():
        return ds.Sequence(length=-1, feature=ds.Value("float32"))

    @staticmethod
    def FloatList():
        return ds.Sequence(length=-1, feature=ds.Value("float32"))

    @staticmethod
    def Imgid():
        return ds.Value("string")

    @staticmethod
    def String():
        return ds.Value("string")

    @staticmethod
    def StringList():
        return ds.Sequence(length=-1, feature=ds.Value("string"))

    @staticmethod
    def NestedStringList():
        return ds.Sequence(ds.Sequence(length=-1, feature=ds.Value("string")))

    @staticmethod
    def Int():
        return ds.Value("int32")

    @staticmethod
    def IntList():
        return ds.Sequence(length=-1, feature=ds.Value("int32"))

    @staticmethod
    def NestedIntList():
        return ds.Sequence(
            length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("int32"))
        )

    @staticmethod
    def Span():
        return ds.Sequence(length=-1, feature=ds.Value("int32"))

    @staticmethod
    def Float():
        return ds.Value("float32")

    @staticmethod
    def Ids():
        return ds.Sequence(length=-1, feature=ds.Value("float32"))

    @staticmethod
    def Boxtensor(n):
        return ds.Array2D((n, 4), dtype="float32")

    # something doesnt look right here (between 2d and 3d features)
    @staticmethod
    def Features2D(d):
        return ds.Array2D((-1, d), dtype="float32")

    @staticmethod
    def Features3D(n, d):
        return ds.Array2D((n, d), dtype="float32")



import inspect
import sys
from collections.abc import Iterable

import torch
import torch.nn.functional as F
import torchvision.transforms.functional as FV
from PIL import Image as PImage
from torchvision.transforms import transforms


def get_scale(obj):
    if not hasattr(obj, "transforms"):
        return None
    scale = None
    for t in obj.transforms:
        if hasattr(t, "_scale"):
            scale = t._scale
    return scale


def get_pad(obj):
    if not hasattr(obj, "transforms"):
        return None
    pad = None
    for t in obj.transforms:
        if hasattr(t, "_pad"):
            pad = t._pad
    return pad


def get_size(obj):
    if not hasattr(obj, "transforms"):
        return None
    size = None
    for t in obj.transforms:
        if hasattr(t, "_size"):
            size = t._size
    return size


def get_rawsize(obj):
    if not hasattr(obj, "transforms"):
        return None
    size = None
    for t in obj.transforms:
        if hasattr(t, "_rawsize"):
            size = t._rawsize
    return size


class FromFile(object):
    _scale = torch.tensor([1.0, 1.0])
    _size = None
    _rawsize = None

    def __init__(self, mode=None, grayscale=False):
        self.mode = mode
        self.grayscale = grayscale
        pass

    def __call__(self, filepath):
        if isinstance(filepath, str):
            if not self.grayscale:
                img = PImage.open(filepath).convert("RGB")
            else:
                img = PImage.open(filepath).convert("L")
            self._size = torch.tensor(img.size)
            self._rawsize = self._size
            return img
        else:
            img = FV.to_pil_image(filepath.byte(), self.mode).convert("RGB")
            return img


class ToTensor(transforms.ToTensor):
    def __call__(self, pil):
        tensor = super().__call__(pil)
        return tensor


class Normalize(transforms.Normalize):
    _std = None
    _mean = None

    def __init__(self, mean=None, std=None, inplace=False):
        super().__init__(mean, std, inplace)

    def __call__(self, tensor):
        # tensor must be: (C, H, W)
        if self.mean is None or self.std is None:
            mean = tensor.mean(dim=(-1, -2))
            std = torch.sqrt((tensor - mean.reshape(-1, 1, 1)) ** 2).mean(dim=(-1, -2))
            mean = mean.tolist()
            std = std.tolist()
            self._std = std
            self._mean = mean
        else:
            return super().__call__(tensor)


# class Pad(object):
#     def __init__(self, size=768, pad_value=0.0):
#         assert isinstance(size, int) or (isinstance(size, Iterable) and len(size) == 2)
#         if isinstance(size, int):
#             max_size = size
#         else:
#             max_size = max(size)
#         self.size = size
#         self.pad_value = pad_value
#         self.max_size = max_size
#         self._size = None

#     @torch.no_grad()
#     def __call__(self, tensor):
#         C, H, W = tensor.shape
#         max_size = self.max_size
#         tensor = F.pad(tensor, [0, max_size - W, 0, max_size - H], value=self.pad_value)
#         self._size = torch.tensor(tensor.shape[1:])
#         return tensor


class Resize(transforms.Resize):
    _size = None
    _rawsize = None
    _scale = None

    def __scale(self):
        if self._size is not None and self._rawsize is not None:
            with torch.no_grad():
                return torch.tensor(
                    [self._size[0] / self._rawsize[0], self._size[1] / self._rawsize[1]]
                )
        else:
            return None

    def __call__(self, pilimg):
        # raise Exception(pilimg.shape)
        self._rawsize = torch.tensor(pilimg.size)
        pilimg = super().__call__(pilimg)
        self._size = torch.tensor(pilimg.size)
        self._scale = self.__scale()
        # raise Exception(self._rawsize, self._size, self._scale)

        return pilimg


class Pad(transforms.Pad):
    # need to implement the amount that the image is padded such that I can resize the
    # binary mask as needed
    _size = None
    _pad = None

    @torch.no_grad()
    def __call__(self, tensor):
        tensor = super().__call__(tensor)
        self._size = torch.tensor(tensor.shape[1:])
        return tensor


class Image:
    def __init__(self):
        if "IMAGEPROCDICT" not in globals():
            global IMAGEPROCDICT
            IMAGEPROCDICT = {
                m[0]: m[1]
                for m in inspect.getmembers(
                    sys.modules["torchvision.transforms.transforms"], inspect.isclass
                )
            }
            IMAGEPROCDICT["ToTensor"] = ToTensor
            IMAGEPROCDICT["FromFile"] = FromFile
            IMAGEPROCDICT["Pad"] = Pad
            IMAGEPROCDICT["Resize"] = Resize
            IMAGEPROCDICT["Normalize"] = Normalize

    def avail(self):
        return list(IMAGEPROCDICT.keys())

    def get(self, name):
        return IMAGEPROCDICT[name]

    def add(self, name, proc):
        IMAGEPROCDICT[name] = proc



import datetime
import os
import random
from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import chain
from typing import Dict, List, Union

import torch
from vltk import COMPLEXPATH
from vltk.abc.visndatasetadapter import VisnDatasetAdapters
from vltk.abc.loop import Loop
from vltk.abc.visnlangdatasetadapter import VisnLangDatasetAdapters
from vltk.inspect import collect_args_to_func, get_classes
from vltk.modeling import Get as Mget
from vltk.modeling.configs import Get

__all__ = ["ComplexExperiment", "ComplexExpIdentifier", "ComplexExperiments"]

_visnlangdatasetadapters = VisnLangDatasetAdapters()
_visndatasetadapters = VisnDatasetAdapters()


class ComplexExperiments:
    def __init__(self):
        if "EXPDICT" not in globals():
            global COMPLEXDICT
            COMPLEXDICT = get_classes(
                COMPLEXPATH, ComplexExpIdentifier, pkg="vltk.complex"
            )

    def avail(self):
        return list(COMPLEXDICT.keys())

    def get(self, name):
        return COMPLEXDICT[name]

    def add(self, name, dset):
        COMPLEXDICT[name] = dset


class ComplexExpIdentifier:
    pass


class ComplexExperiment(ComplexExpIdentifier, ABC):
    # for now lets just define dataset in loop
    def __init__(self, config, datasets):

        # maybe we can assume experiments are homegeneous within each loop
        # defualt stuff
        self.cur_epoch = 0
        self.datasets = datasets
        self.config = config
        self.seed = self.config.seed
        random.seed(self.seed)
        torch.manual_seed(self.seed)
        assert self.datasets is not None, "must specify 'datasets' when firing command"
        self.epochs = self.config.train.epochs
        self.logdir = getattr(self.config, "logdir", None)
        if self.logdir is not None and self.config.logging:
            os.makedirs(self.logdir, exist_ok=True)

        self._init_datasets()
        self._init_models()
        self._init_loops()
        self._init_gradient_tracking()
        self._experiment_outputs = {}

    @property
    def experiment_outputs(self):
        return self._experiment_outputs

    @property
    def model_dict(self):
        return self._model_dict

    @property
    def loop_dict(self):
        return self._loop_dict

    @property
    def loop_info(self):
        return self._loop_dict

    @property
    def loop_order(self):
        return self._order

    @property
    def extra_modules(self):
        return None

    def get_model(self):
        pass

    def set_model_devices(self):
        if self.model_dict:
            for name, model in self.model_dict.items():
                if name not in self.config.models.main_model:
                    model = model.to(torch.device(self.config.aux_gpu))
                else:
                    model = model.to(torch.device(self.config.gpu))
        # set all extra torch.nn.Modules to main gpu for now
        if self.extra_modules is not None:
            for name, nn in self.extra_modules.items():
                nn = nn.to(torch.device(self.config.gpu))

    def _init_gradient_tracking(self):
        # hardcoded now for time
        if self.model_dict:
            for name, model in self.model_dict.items():
                name = name.split("_")[0]
                conf = getattr(self.config.models, name)
                if hasattr(conf, "freeze_layers"):
                    layers_to_freeze = conf.freeze_layers
                    for n, p in model.named_parameters():
                        if "layer" in n.lower():
                            if layers_to_freeze.pop(0):
                                p.requires_grad = False

    def _init_datasets(self):
        # first check to see if any train or any val
        any_train = False
        any_val = False
        # check for train and eval loops
        for loop_key in self.loops_to_models:
            loop_cls = loop_key
            if loop_cls.is_train:
                any_train = True
            else:
                any_val = True

        self.label_to_id = {}
        self.dataset2splits = defaultdict(set)
        self.train_visnlangdatasetadapterdict = defaultdict(dict)
        self.eval_visnlangdatasetadapterdict = defaultdict(dict)
        self.train_visndatasetadapterdict = defaultdict(dict)
        self.eval_visndatasetadapterdict = defaultdict(dict)
        train_datasets = set()
        eval_datasets = set()
        train_splits = set()
        eval_splits = set()

        # loop through train datasets
        for (dset, splits) in self.datasets:
            train_datasets.add(dset)
            train_splits = train_splits.union(splits)
            self.dataset2splits[dset] = self.dataset2splits[dset].union(splits)
        # then loop thorugh eval datasets
        for (dset, splits) in self.config.data.eval_datasets:
            assert (
                dset in train_datasets
            ), "eval datasets must also be present in train datasets"
            eval_datasets.add(dset)
            eval_splits = eval_splits.union(splits)
            if not splits.intersection(self.dataset2splits[dset]):
                self.dataset2splits[dset] = self.dataset2splits[dset].union(splits)

        label_id = 0
        for name in sorted(set(self.dataset2splits.keys())):
            for split in sorted(set(self.dataset2splits[name])):
                if (
                    name in eval_datasets
                    and split in eval_splits
                    and not self.config.data.skip_eval
                    and any_val
                ) or (
                    split in train_splits
                    and not self.config.data.skip_train
                    and any_train
                ):
                    visnlangdatasetadapter = _visnlangdatasetadapters.get(name).from_config(
                        self.config.data, splits=split
                    )[split]
                else:
                    continue
                for l in sorted(visnlangdatasetadapter.labels):
                    if l not in self.label_to_id:
                        self.label_to_id[l] = label_id
                        label_id += 1
                print(f"Added VisnLangDatasetAdapter {name}: {split}")
                if (
                    name in eval_datasets
                    and split in eval_splits
                    and not self.config.data.skip_eval
                    and any_val
                ):
                    self.eval_visnlangdatasetadapterdict[name][split] = visnlangdatasetadapter
                if (
                    split in train_splits
                    and not self.config.data.skip_train
                    and any_train
                ):
                    self.train_visnlangdatasetadapterdict[name][split] = visnlangdatasetadapter
                is_name, is_split = zip(*visnlangdatasetadapter.data_info[split].items())
                is_name = is_name[0]
                is_split = is_split[0][0]
                if self.config.data.extractor is not None:
                    is_path = visnlangdatasetadapter.get_arrow_split(
                        self.config.data.datadirs, is_split, self.config.data.extractor
                    )
                    visndatasetadapter = _visndatasetadapters.get(self.config.data.extractor).from_file(
                        is_path
                    )
                else:
                    visndatasetadapter = visnlangdatasetadapter.get_imgid_to_raw_path(
                        self.config.data.datadirs, is_split
                    )

                print(f"Added VisnDatasetAdapter {is_name}: {is_split}")

                if (
                    name in eval_datasets
                    and split in eval_splits
                    and not self.config.data.skip_eval
                    and any_val
                ):
                    self.eval_visndatasetadapterdict[is_name][is_split] = visndatasetadapter
                if (
                    split in train_splits
                    and not self.config.data.skip_train
                    and any_train
                ):
                    self.train_visndatasetadapterdict[is_name][is_split] = visndatasetadapter

    def _init_models(self):
        model_dict = {}
        models = set(chain(*list(self.loops_to_models.values())))
        for model in models:
            if not isinstance(model, str):
                raise Exception
            # get model class
            mclass = Mget[model]
            # get modl config
            mconfig = Get[model.split("_")[0]]

            # add model config to the vltk config object
            self.config.models.add(model.split("_")[0], mconfig)
            # get pointer to model config in vltk config
            config = getattr(self.config.models, model.split("_")[0])
            # instantiate model
            if getattr(config, "checkpoint", None) is not None and hasattr(
                mclass, "from_pretrained"
            ):
                model_instance = mclass.from_pretrained(
                    config.checkpoint, config=config
                )
            elif not hasattr(mclass, "from_pretrained"):
                # mandatory = false: ignore all required args. however we need to instead
                # get something as to where we get as many required args as are in the dict as possible
                arg_dict = collect_args_to_func(
                    mclass.__init__, config.to_dict(), mandatory=False
                )
                arg_dict["config"] = config
                model_instance = mclass(**arg_dict)
                if getattr(config, "checkpoint", None) is not None:
                    model_instance.load_state_dict(
                        torch.load(config.checkpoint), strict=False
                    )

            model_dict[model] = model_instance
        self._model_dict = model_dict

    def _init_loops(self):

        loop_dict = {}
        loop_info = {}
        for loop_key in self.loops_to_models:
            assert isinstance(loop_key, Loop)
            loop_cls = loop_key
            loop_name = loop_cls.name
            is_train = loop_cls.is_train
            if is_train:
                if self.config.data.skip_train:
                    continue
                visnlangdatasetadapterdict = self.train_visnlangdatasetadapterdict
                visndatasetadapterdict = self.train_visndatasetadapterdict
            else:
                visnlangdatasetadapterdict = self.eval_visnlangdatasetadapterdict
                visndatasetadapterdict = self.eval_visndatasetadapterdict
                if self.config.data.skip_eval or len(visnlangdatasetadapterdict) == 0:
                    continue

            loop = loop_cls(
                config=self.config,
                model_dict={
                    k: v
                    for k, v in self.model_dict.items()
                    if k in self.loops_to_models.get(loop_key)
                },
                extra_modules=self.extra_modules,
                datasets=self.datasets,
                visndatasetadapterdict=visndatasetadapterdict,
                visnlangdatasetadapterdict=visnlangdatasetadapterdict,
                label_dict=self.label_to_id,
            )

            if (loop.is_train and not self.config.data.skip_train) or (
                not loop.is_train and not self.config.data.skip_eval
            ):
                loop_dict[loop_name] = loop
                loop_info[loop_name] = loop.is_train

        print(f"Loaded Loops: {list(loop_dict.keys())}")

        self._loop_dict = loop_dict
        self._loop_info = loop_info
        order = sorted(
            list(loop_info.keys()), key=lambda x: int(loop_info[x]), reverse=True
        )
        self._order = order

    def write(self, epoch_output: dict = None):
        info = self.loginfo(epoch_output)
        if self.config.logging and info is not None and info:
            logfile = os.path.join(self.config.logdir, "log.txt")
            assert logfile is not None
            with open(logfile, "a") as f:
                date = datetime.datetime.now()
                f.write(f"Time: {date} \n {info} \n")
                f.flush()
            return True
        return False

    def write_iter(self, info: dict = None):
        if info is not None:
            clean_info = {}
            for k, v in info.items():
                if v is None or (not isinstance(v, bool) and not v):
                    continue
                else:
                    clean_info[k] = v
        logstr = ""
        for k, v in info.items():
            logstr += f"{k}={v}; "

        if self.config.logging and info is not None and info:
            logfile = os.path.join(self.config.logdir, "cur_epoch.txt")
            assert logfile is not None
            with open(logfile, "a") as f:
                date = datetime.datetime.now()
                f.write(f"Time: {date} \n {info} \n")
                f.flush()
            return True
        return False

    def save(self):
        print("\nsaving...\n")
        if self.model_dict:
            for name, model in self.model_dict.items():
                save_name = name + f"_{self.cur_epoch}.pt"
                save_name = os.path.join(self.config.logdir, save_name)
                torch.save(model.state_dict(), save_name)
        if self.extra_modules is not None:
            for extra, torch_module in self.extra_modules.items():
                save_name = extra + f"_{self.cur_epoch}.pt"
                save_name = os.path.join(self.config.logdir, save_name)
                torch.save(torch_module.state_dict(), save_name)
        optim_dict = {}
        save_name = f"optims_{self.cur_epoch}.pt"
        save_name = os.path.join(self.config.logdir, save_name)
        for loop_name, loop in self:
            if loop.is_train:
                optim_dict[loop_name] = loop.optim.state_dict()
        torch.save(optim_dict, save_name)
        save_name = os.path.join(self.config.logdir, "exp_outputs.pkl")
        """
        TODO: FIX
        self.experiment_outputs.dump(save_name)
        """
        save_name = os.path.join(self.config.logdir, "config.yaml")
        self.config.dump_yaml(save_name)

    def get_exp_info(self):
        exp_info = {
            "name": self.name,
            "datasets": self.datasets,
            "loops_to_models": self.loops_to_models,
            "cur_steps": {},
            "cur_epoch": self.cur_epoch,
            "epochs": self.epochs,
            "total_steps": {},
            "schedulers": {},
            "warmups": {},
        }
        for loop_name, loop in self:
            exp_info["cur_steps"][loop_name] = loop.cur_step
            exp_info["total_steps"][loop_name] = loop.total_steps
            if loop.is_train and loop.warmup is not None:
                exp_info["warmups"][loop_name] = loop.warmup.state_dict()

        return exp_info

    def __iter__(self):
        for loop_name in self.loop_order:
            yield loop_name, self.loop_dict.get(loop_name)

    def __call__(self):
        print()
        self.set_model_devices()
        for epoch in range(self.epochs):
            epoch_output = {}
            self.cur_epoch = epoch
            any_train = False
            for i, (loop_name, loop) in enumerate(self):
                loop_output = loop()
                epoch_output[loop_name] = loop_output
                if loop.is_train:
                    any_train = True

                if len(self.loop_order) - 1 == i:
                    break

            self.experiment_outputs[f"epoch_{epoch}"] = epoch_output
            self.write(loop_output)
            if self.config.test_save or (self.config.save_after_epoch and any_train):
                self.save()
        if (
            not self.config.save_after_epoch
            and not self.config.test_run
            and ((any_train and self.config.save_after_exp) or self.config.test_save)
        ):
            self.save()

    @property
    @abstractmethod
    def name(self) -> str:
        return ""

    @property
    @abstractmethod
    def loops_to_models(self) -> Dict[Union[str, object], List[str]]:
        return {}

    @abstractmethod
    def loginfo(self, **kwargs) -> str:
        return ""



import inspect
# note if we do not immport a pacakage correctly in this class, no loops or exps will be present
import json
import math
import os
import random
import resource
import sys
from collections import Iterable

import torch
import vltk.vars as vltk
# disable logging from datasets
from datasets.utils.logging import set_verbosity_error
from vltk.dataset.basedataset import BaseDataset, CollatedVLSets
from vltk.processing import LangProcessor, Processors, VisnLangProcessor

__import__("tokenizers")
TOKENIZERS = {
    m[0]: m[1]
    for m in inspect.getmembers(sys.modules["tokenizers"], inspect.isclass)
    if "Tokenizer" in m[0]
}

rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (6144, rlimit[1]))

set_verbosity_error()


os.environ["TOKENIZERS_PARALLELISM"] = "False"


class LangDataset(BaseDataset):
    def __init__(
        self,
    ):
        super().__init()
        """
        Implementation for LangDataset is not completed or really developed at all yet.

        This serves as a placeholder class for all relevant pure-text
        processing in the VisnLang Dataset so that I can have that class inherit from
        VisionDataset + LangDataset while I perform this refactoring
        """

    def _init_lang_processors(self, config):
        lang_processors = (
            config.lang_processors if config.lang_processors is not None else []
        )

        lang_processors = [
            x if not isinstance(x, str) else Processors().get(x)
            for x in lang_processors
        ]

        lang_processors = list(
            filter(
                lambda x: x.__bases__[0] == LangProcessor,
                lang_processors,
            )
        )

        self.lang_processors = [
            x(
                tokenizer=self.tokenizer,
                from_transformers=self.from_transformers,
                config=self.config,
                metadata_ids=self.metadata_ids,
            )
            for x in lang_processors
        ]

        self.lang_processor_keys = ()
        for x in self.lang_processor_keys:
            self.lang_processor_keys += tuple(x.keys)

    def run_lang_processors(self, entry, encode_batch=False):
        proc_args = self.lang_processor_args()
        if not encode_batch:
            for processor in self.lang_processors:
                entry = processor(entry, **proc_args)
        else:
            # add ability later to allow this to happen in parallel
            keys, values = zip(*entry.items())
            new_keys = keys
            values_t = list(zip(*values))
            for i, v in enumerate(values_t):
                e = dict(zip(keys, v))
                for processor in self.lang_processors:
                    e = processor(e, **proc_args)
                new_keys, temp_values = zip(*e.items())
                values_t[i] = list(temp_values)
            new_values = list(zip(*values_t))
            entry = dict(zip(new_keys, new_values))
        return entry

    def lang_processor_args(self):
        # max_rand_sents = 1 if not self.config.img_first else 32
        return {
            "tokenizer": self.tokenizer,
            "config": self.config,
            # "random_sents": [self.random_sent() for i in range(max_rand_sents)],
            "special_ids": self.special_ids,
            "metadata_ids": self.metadata_ids,
            "all_ids": self.all_ids,
            "n_ids": len(self.all_ids),
        }

    def tokenize_entry(self, x, encode_batch=False):
        tokenizer = self.tokenizer
        proc_args = self.lang_processor_args()
        from_transformers = self.from_transformers
        if not from_transformers and not encode_batch:
            encoded = tokenizer.encode(x.pop(vltk.text))
            x[vltk.text_attention_mask] = encoded.attention_mask
            x[vltk.input_ids] = encoded.ids
            x[vltk.type_ids] = encoded.type_ids
        elif from_transformers and not encode_batch:
            encoded = tokenizer(
                x.pop(vltk.text),
                padding="max_length",
                truncation="longest_first",
                max_length=proc_args["config"].lang.max_seq_length,
                return_token_type_ids=True,
            )
            x[vltk.text_attention_mask] = encoded["attention_mask"]
            x[vltk.input_ids] = encoded["input_ids"]
            x[vltk.type_ids] = encoded["token_type_ids"]
        elif not from_transformers and encode_batch:
            encoded = tokenizer.encode_batch(x.pop(vltk.text))
            x[vltk.text_attention_mask] = torch.tensor(
                list(map(lambda x: x.attention_mask, encoded))
            )
            x[vltk.input_ids] = torch.tensor(list(map(lambda x: x.ids, encoded)))
            x[vltk.type_ids] = torch.tensor(list(map(lambda x: x.type_ids, encoded)))
        else:
            text = x.pop(vltk.text)
            encoded = tokenizer(
                list(text),
                padding="max_length",
                truncation="longest_first",
                max_length=proc_args["config"].lang.max_seq_length,
                return_token_type_ids=True,
            )
            x[vltk.text_attention_mask] = torch.tensor(encoded["attention_mask"])
            x[vltk.input_ids] = torch.tensor(encoded["input_ids"])
            x[vltk.type_ids] = torch.tensor(encoded["token_type_ids"])
        return x

    def __len__(self):
        return int(math.floor(len(self.datasets) * self.config.percent))

    def random_sent(self):
        rand_ind = random.randint(0, len(self.datasets) - 1)
        text_info = self.datasets[rand_ind]
        rand_sent = text_info[vltk.text]
        return rand_sent

    def _handle_text_label(self, entry, encode_batch=False):
        if vltk.label in entry and not encode_batch:
            label = entry.pop(vltk.label)
            if isinstance(label, torch.Tensor):
                entry[vltk.label] = label
                return entry
            entry[vltk.label] = torch.tensor(self.metadata_ids[vltk.label][label[0]])
            if vltk.score in entry:
                entry[vltk.score] = torch.tensor(entry[vltk.score])
            elif vltk.score in self.max_spanning_cols:
                entry[vltk.score] = torch.ones(entry[vltk.label].shape)
        elif (
            vltk.label not in entry
            and not encode_batch
            and vltk.label in self.max_spanning_cols
        ):
            # this condition is for an entry without label but label is in one dataset
            entry[vltk.label] = torch.tensor(self.config.lang.ignore_id)
            if vltk.score in self.max_spanning_cols:
                entry[vltk.score] = torch.tensor(0).float()

        elif (
            vltk.label not in entry
            and encode_batch
            and vltk.label in self.max_spanning_cols
        ):
            # this condition is for an entry without label but label is in one dataset
            # except across full batch
            entry[vltk.label] = torch.tensor(
                [self.config.lang.ignore_id] * entry[vltk.input_ids].shape[0]
            )
            if vltk.score in self.max_spanning_cols:
                entry[vltk.score] = torch.zeros(entry[vltk.input_ids].shape[0]).float()

        elif (
            vltk.label in entry
            and encode_batch
            and vltk.label in self.max_spanning_cols
        ):
            # TODO dont worry about mulitple labels per thing just yet

            label = entry.pop(vltk.label)
            if isinstance(label, torch.Tensor):
                entry[vltk.label] = label
                return entry
            lids = []
            for l in label:
                lid = self.metadata_ids[vltk.label][l[0]]
                lids.append(lid)
            entry[vltk.label] = torch.tensor(lids)
            if vltk.score in entry:
                entry[vltk.score] = torch.tensor([[v[0] for v in entry[vltk.score]]])
            elif vltk.score in self.max_spanning_cols:
                entry[vltk.score] = torch.ones(entry[vltk.label].shape)
        else:
            pass
        return entry

    def _handle_text_annotations(self, entry, encode_batch=False):
        entry = self.run_lang_processors(entry, encode_batch=encode_batch)

        entry = self.tokenize_entry(entry, encode_batch=encode_batch)
        entry = self._handle_text_label(entry, encode_batch=encode_batch)
        return entry



import io
import os
import unittest

import numpy as np
from PIL import Image

from vltk import SingleImageViz


PATH = os.path.dirname(os.path.realpath(__file__))

URL = "https://raw.githubusercontent.com/airsplay/py-bottom-up-attention/master/demo/data/images/input.jpg"


class TestVisaulizer(unittest.TestCase):

    url = URL

    def test_viz(self):
        viz = SingleImageViz(self.url)
        viz.show()



from collections import defaultdict


from vltk.features import Features
from vltk import adapters
import vltk.vars as vltk


class Coco2014(adapters.VisnDataset):
    @staticmethod
    def schema():
        return {
            vltk.box: Features.Box(),
            vltk.polygons: Features.Polygons(),
            vltk.objects: Features.StringList(),
        }

    @staticmethod
    def forward(json_files, splits):

        total_annos = {}
        id_to_cat = {}
        file_to_id_to_stem = defaultdict(dict)
        for file, json in json_files.items():
            if "instance" not in file:
                continue
            info = json["images"]
            for i in info:
                img_id = i["file_name"].split(".")[0]
                file_to_id_to_stem[file][i["id"]] = img_id
        for file, json in json_files.items():
            if "instance" not in file:
                continue

            categories = json["categories"]
            for cat in categories:
                id_to_cat[cat["id"]] = cat["name"]

            for entry in json["annotations"]:
                # TODO: change this image ID thing later

                img_id = str(file_to_id_to_stem[file][entry["image_id"]])
                bbox = entry["bbox"]
                segmentation = entry["segmentation"]
                category_id = id_to_cat[entry["category_id"]]
                if entry["iscrowd"]:
                    seg_mask = []
                else:
                    seg_mask = segmentation
                    if not isinstance(seg_mask[0], list):
                        seg_mask = [seg_mask]
                img_data = total_annos.get(img_id, None)
                if img_data is None:
                    img_entry = defaultdict(list)
                    img_entry[vltk.objects].append(category_id)
                    img_entry[vltk.box].append(bbox)
                    img_entry[vltk.polygons].append(seg_mask)
                    total_annos[img_id] = img_entry
                else:
                    total_annos[img_id][vltk.box].append(bbox)
                    total_annos[img_id][vltk.objects].append(category_id)
                    total_annos[img_id][vltk.polygons].append(seg_mask)

        return [{vltk.imgid: img_id, **entry} for img_id, entry in total_annos.items()]



import inspect
import os
import sys
from abc import abstractmethod
from collections import OrderedDict

import datasets as ds
import pyarrow
import torch
import vltk.vars as vltk
from datasets import ArrowWriter
from tqdm import tqdm
from vltk.abc.adapter import Adapter
from vltk.configs import VisionConfig
from vltk.inspection import collect_args_to_func
from vltk.processing.image import get_rawsize, get_scale, get_size


class VisnExtraction(Adapter):
    _meta_names = [
        "img_to_row_map",
        "dataset",
        "processor_args",
    ]
    _is_feature = True
    _batch_size = 128

    default_processor = None

    def processor(self, *args, **kwargs):
        return self._processor(*args, **kwargs)

    def align_imgids(self):
        for i in range(len(self)):
            self._img_to_row_map[self[i][vltk.imgid]] = i

    def check_imgid_alignment(self):
        orig_map = self.img_to_row_map
        for i in range(len(self)):
            img_id = self[i][vltk.imgid]
            mapped_ind = orig_map[img_id]
            if mapped_ind != i:
                return False
            self._img_to_row_map[self[i][vltk.imgid]] = i
        return True

    @property
    def processor_args(self):
        return self._processor_args

    @property
    def dataset(self):
        return self._dataset.decode()

    @property
    def config(self):
        return self._config

    @staticmethod
    def _check_forward(image_preprocessor, forward):
        pass
        args = str(inspect.formatargspec(*inspect.getargspec(forward)))
        assert "entry" in args, (args, type(args))
        assert "model" in args, (args, type(args))
        assert callable(image_preprocessor), (
            image_preprocessor,
            callable(image_preprocessor),
        )

    @staticmethod
    def _build_image_processor(config, processor_class, default_processor):
        if config is None:
            processor_args = {}
        else:
            if isinstance(config, dict):
                processor_args = config
            else:
                processor_args = config.to_dict()
        if processor_class is not None:
            processor = processor_class(**processor_args)
        elif config is not None:
            processor = config.build()
        elif config is None:
            if default_processor is None:
                processor_class = VisionConfig()
                processor = processor_class.build()
            else:
                processor_class = default_processor
                processor_args = default_processor.to_dict()
            processor = processor_class.build()

        return processor, processor_args

    @classmethod
    def extract(
        cls,
        datadir,
        processor_config=None,
        splits=None,
        subset_ids=None,
        dataset=None,
        img_format="jpg",
        processor=None,
        **kwargs,
    ):

        dataset_name = dataset
        searchdir = datadir
        extractor_name = cls.__name__.lower()
        searchdirs, valid_splits = cls._get_valid_search_pathes(
            searchdir, dataset_name, splits
        )
        savedir = VisnExtraction._make_save_path(
            searchdir, dataset_name, extractor_name
        )
        processor, processor_args = VisnExtraction._build_image_processor(
            processor_config, processor, cls.default_processor
        )
        schema = VisnExtraction._build_schema(cls.schema, **kwargs)

        try:
            model, model_config = cls.setup()
        except Exception:
            raise Exception(
                "setup model is supposed to return (`model`, `model_config`) objects, but only returned one object."
            )
        setattr(cls, "model", model)
        # setup tracking dicts
        split2buffer = OrderedDict()
        split2stream = OrderedDict()
        split2writer = OrderedDict()
        split2imgid2row = {}
        split2currow = {}
        split2metadata = {}
        # begin search
        print(f"extracting from {searchdirs}")
        batch_size = cls._batch_size
        cur_size = 0
        cur_batch = None
        files = set(cls._iter_files(searchdirs, iter_imgs=True))
        total_files = len(files)
        for i, path in tqdm(
            enumerate(files),
            file=sys.stdout,
            total=total_files,
        ):
            path_list = path.split("/")
            split = path_list[-2]
            img_id = path_list[-1].split(".")[0]
            imgs_left = abs(i + 1 - total_files)
            if split not in valid_splits:
                continue
            if subset_ids is not None and img_id not in subset_ids:
                continue

            # oragnize by split now
            schema = ds.Features(schema)
            if split not in split2buffer:
                meta_dict = VisnExtraction._init_metadata(schema)
                imgid2row = {}
                cur_row = 0
                cur_size = 0
                buffer = pyarrow.BufferOutputStream()
                split2buffer[split] = buffer
                stream = pyarrow.output_stream(buffer)
                split2stream[split] = stream
                writer = ArrowWriter(features=schema, stream=stream)
                split2writer[split] = writer
                split2metadata[split] = meta_dict
            else:
                # if new split and cur size is not zero, make sure to clear
                if cur_size != 0:
                    cur_size = 0
                    batch = schema.encode_batch(cur_batch)
                    writer.write_batch(batch)
                meta_dict = split2metadata[split]
                imgid2row = split2imgid2row[split]
                cur_row = split2currow[split]
                buffer = split2buffer[split]
                stream = split2stream[split]
                writer = split2writer[split]

            if img_id in imgid2row:
                print(f"skipping {img_id}. Already written to table")
            imgid2row[img_id] = cur_row
            cur_row += 1
            split2currow[split] = cur_row
            split2imgid2row[split] = imgid2row
            filepath = str(path)

            entry = {vltk.filepath: filepath, vltk.imgid: img_id, vltk.split: split}
            entry[vltk.img] = processor(filepath)
            entry[vltk.size] = get_size(processor)
            entry[vltk.scale] = get_scale(processor)
            entry[vltk.rawsize] = get_rawsize(processor)
            # now do model forward

            forward_dict = collect_args_to_func(cls.forward, kwargs=kwargs)
            output_dict = cls.forward(model=model, entry=entry, **forward_dict)
            assert isinstance(
                output_dict, dict
            ), "model outputs should be in dict format"
            output_dict[vltk.imgid] = [img_id]
            meta_dict = VisnExtraction._update_metadata(meta_dict, output_dict)

            if cur_size == 0:
                cur_batch = output_dict
                cur_size = 1
            else:
                # TODO: check if this is right
                for k, v in output_dict.items():
                    cur_batch[k].extend(v)
                    cur_size += 1

            # write features
            if cur_size == batch_size or imgs_left < batch_size:
                cur_size = 0
                batch = schema.encode_batch(cur_batch)
                writer.write_batch(batch)
            split2imgid2row[split] = imgid2row

        # define datasets
        splitdict = {}
        meta_dict = {}
        print("saving...")
        for (_, writer), (split, b) in zip(split2writer.items(), split2buffer.items()):
            savefile = os.path.join(savedir, f"{split}.arrow")
            imgid2row = split2imgid2row[split]
            meta_dict = split2metadata[split]
            meta_dict["img_to_row_map"] = imgid2row
            meta_dict["model_config"] = model_config
            meta_dict["dataset"] = dataset
            meta_dict["processor_args"] = processor_args

            table, info, meta_dict = VisnExtraction._save_dataset(
                b, writer, savefile, meta_dict, split
            )

            # return class
            arrow_dset = cls(
                arrow_table=table,
                split=split,
                info=info,
                meta_dict=meta_dict,
            )
            splitdict[split] = arrow_dset

        return splitdict

    @abstractmethod
    def forward(model, entry, **kwargs):
        raise Exception("child forward is not being called")

    @abstractmethod
    def schema(*args, **kwargs):
        return dict

    @abstractmethod
    def setup():
        return None



import importlib
import inspect
import os
import sys

import torch.nn as nn

PATH = os.path.join(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..")), "libdata"
)


def get_args(func, dictionary):
    argspec = inspect.getfullargspec(func)
    args = argspec.args
    kwargs = 0 if argspec.varkw is None else 1
    if kwargs:
        return dictionary
    elif args:
        argdict = {}
        for a in args:
            if a in dictionary:
                argdict[a] = dictionary[a]
        return argdict
    else:
        return None


def get_classes(path_or_dir_name, cls_defintion=None, pkg=None):
    os.chdir(os.getcwd())
    # for single file
    if os.path.isfile(path_or_dir_name):
        clsses = import_classes_from_file(path_or_dir_name, pkg=pkg)
        filter_dict = {}
        if cls_defintion is not None:
            for n, c in clsses.items():
                if cls_defintion in inspect.getmro(c) or hasattr(c, "name"):
                    try:
                        filter_dict[c.name] = c
                    except Exception:
                        filter_dict[c.__name__.lower()] = c
            return filter_dict
        else:
            return clsses

    # for dir
    classes = {}

    sys.path.insert(0, path_or_dir_name)
    module_name = None
    for p in os.listdir(path_or_dir_name):
        if p[0] != "_":
            if pkg is None and p[-3:] == ".py":
                module_name = p[:-3]
                npkg = __import__(module_name, fromlist=[""])
            elif pkg is None and p[-3:] != ".py":
                continue
            else:
                npkg = pkg + f".{p.split('.')[0]}"
            try:
                if pkg is None:
                    mod = inspect.getmembers(npkg, inspect.isclass)
                    # if module_name == 'exp_newdata':
                    #    raise Exception(mod)
                else:
                    try:
                        mod = importlib.import_module(npkg)
                    except Exception:
                        mod = importlib.import_module(npkg)
                        # raise Exception(p, npkg, mod)
                    mod = inspect.getmembers(mod, inspect.isclass)
                for t in mod:
                    if cls_defintion in inspect.getmro(t[-1]):
                        # print(t[-1].__abstractmethods__)
                        if not inspect.isabstract(t[-1]):
                            try:
                                classes[t[-1].__name__.lower()] = t[-1]
                            except AttributeError:
                                classes[t[-1].__name__.lower()] = t[-1]
            except Exception:
                mod = importlib.import_module(npkg)
                assert module_name != "exp_newdata"
                pass
    return classes


def get_func_signature_v2(func):
    required = set()
    keyword = {}
    sig = inspect.signature(func).parameters
    for k, v in sig.items():
        if v.default == inspect._empty:
            required.add(k)
        else:
            keyword[k] = v.default
    return required, keyword


def collect_args_to_func(func, kwargs=None, mandatory=False):
    func_input = {}
    if kwargs is None:
        kwargs = {}
    else:
        assert isinstance(kwargs, dict)
    req, keyw = get_func_signature_v2(func)
    if mandatory:
        for r in req:
            assert r in kwargs, (
                "\n"
                f"The required args of {func.__name__} are: {req}"
                f" but '{r}' not found in kwargs: {list(kwargs.keys())}"
            )
            func_input[r] = kwargs[r]
    for k in keyw:
        if k in kwargs:
            func_input[k] = kwargs[k]
    return func_input


def apply_args_to_func(func, kwargs=None, mandatory=True):
    func_input = {}
    if kwargs is None:
        kwargs = {}
    else:
        assert isinstance(kwargs, dict)
    req, keyw = get_func_signature_v2(func)
    if mandatory:
        for r in req:
            assert r in kwargs, (
                "\n"
                f"The required args of {func.__name__} are: {req}"
                f" but '{r}' not found in kwargs: {list(kwargs.keys())}"
            )
            func_input[r] = kwargs[r]
    for k in keyw:
        if k in kwargs:
            func_input[k] = kwargs[k]
    return func(**func_input)


def is_cls(inspect_class, cls_defintion):
    if (
        cls_defintion in inspect.getmro(inspect_class)
        and inspect_class.__name__ != cls_defintion.__name__
    ):
        return True
    else:
        return False


def is_model(inspect_class):
    if (
        nn.Module in inspect.getmro(inspect_class)
        and inspect_class.__name__ != "Module"
    ):
        return True
    else:
        return False


def import_from_dir(clsdir, pkg):
    modules = []
    for p in os.listdir(clsdir):
        file = os.path.join(clsdir, p)
        mods = import_classes_from_file(file, pkg)
        modules.extend(mods)

    return modules


def import_classes_from_file(clspath, pkg=None):
    clsfile = clspath.split("/")[-1]
    clsname = clsfile.split(".")[0]
    if pkg is not None:
        clsname = pkg + f".{clsname}"
    clsdir = clspath.replace(clsfile, "")
    sys.path.insert(0, clsdir)
    mod = importlib.import_module(clsname, package=None)
    return {clss[0]: clss[1] for clss in inspect.getmembers(mod, inspect.isclass)}


def import_funcs_from_file(clspath, pkg):
    clsfile = clspath.split("/")[-1]
    clsname = clsfile.split(".")[0]
    if pkg is not None:
        clsname = pkg + f".{clsname}"
    clsdir = clspath.replace(clsfile, "")
    sys.path.insert(0, clsdir)
    mod = importlib.import_module(clsname, package=pkg)
    return {func[0]: func[1] for func in inspect.getmembers(mod, inspect.isfunction)}


def get_func_signature(func):
    sig = inspect.signature(func).parameters
    return sig



import os
import unittest

import torch
from vltk import Config, GeneralizedRCNN, Preprocess, SingleImageViz, get_data

PATH = os.path.dirname(os.path.realpath(__file__))
URL = "https://raw.githubusercontent.com/airsplay/py-bottom-up-attention/master/demo/data/images/input.jpg"


class TestFRCNNForwardPass(unittest.TestCase):
    def test_forward(self):
        # load models and model components
        # frcnn_cfg = Config.from_pretrained(f"{PATH}/config.yaml")
        frcnn = GeneralizedRCNN.from_pretrained("unc-nlp/frcnn-vg-finetuned")
        frcnn.roi_outputs.nms_thresh = [0.5, 1.0, 0.1]
        frcnn.roi_outputs.score_thresh = 0.2
        frcnn.roi_outputs.min_detections = 36
        frcnn.roi_outputs.max_detections = 36
        frcnn_cfg = frcnn.config
        # Run the actual model
        image_preprocess = Preprocess(frcnn_cfg)
        images, sizes, scales_yx = image_preprocess(URL)
        frcnn(
            images,
            sizes,
            scales_yx=scales_yx,
            padding="max_detections",
            max_detections=frcnn_cfg.max_detections,
            return_tensors="np",
        )


if __name__ == "__main__":
    Image = SingleImageViz(
        URL,
        id2obj=get_data("./tests/objects.txt"),
        id2attr=get_data("./tests/attributes.txt"),
    )

    # load models and model components
    frcnn = GeneralizedRCNN.from_pretrained("unc-nlp/frcnn-vg-finetuned")
    frcnn.roi_outputs.min_detections = 36
    frcnn.roi_outputs.max_detections = 36
    # Run the actual model
    image_preprocess = Preprocess(frcnn.config)
    images, sizes, scales_yx = image_preprocess(URL)
    output_dict = frcnn(
        images,
        sizes,
        scales_yx=scales_yx,
        padding="max_detections",
        max_detections=frcnn.config.max_detections,
        return_tensors="np",
    )

    Image.draw_boxes(
        output_dict["boxes"],
        output_dict["obj_ids"],
        output_dict["obj_probs"],
        output_dict["attr_ids"],
        output_dict["attr_probs"],
    )
    Image.save('test.jpg')



from transformers import RobertaTokenizerFast
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig
from vltk.dataset.builder import init_datasets
from vltk.processing import LangProcessor


class TestProcessor(LangProcessor):
    def forward(self, x, *args, **kwargs):
        return x


if __name__ == "__main__":

    datadir = "/home/eltoto/demodata"
    config = DataConfig(
        train_datasets=[
            ["gqa", "train"],
            ["vqa", "trainval"],
            ["cococaptions", "trainval"],
            ["vgqa", "train"],
        ],
        extractor="frcnn",
        datadir=datadir,
        num_workers=0,
        train_batch_size=2,
        img_first=True,
        ignore_segmentation=True,
    )
    train, val = init_datasets(config)
    for i, b in enumerate(train):
        if i == 0:
            continue
        b = train.transpose_vl(b)
        break



from collections import Counter

import vltk.vars as vltk
from vltk import adapters
from vltk.adapters import Adapters
from vltk.features import Features
from vltk.utils.adapters import clean_label


class VGQA(adapters.VisnDataset):
    data_info = {
        "train": {"visualgenome": ["train"]},
    }

    def schema():
        # img id and score are assumed to be default features
        return {
            vltk.qid: Features.String(),
            vltk.label: Features.StringList(),
        }

    def adjust_imgid(imgid, vdset_name, vdset_split):
        # length of COCO ids are are always length 12
        # imgid = f'{"COCO"}_{vdset_split[0].lower()}{2014}_{"".join(["0"] * (12 - len(imgid)))}{imgid}'
        return imgid

    def forward(json_files, split, min_label_frequency=9):
        batch_entries = []
        answer_counts = Counter()
        skipped = 0
        for filename, data in json_files.items():
            for y in data:
                for x in y["qas"]:
                    answer_counts.update([clean_label(x["answer"])]),

        for filename, data in json_files.items():
            for y in data:
                for x in y["qas"]:
                    if answer_counts[clean_label(x["answer"])] >= min_label_frequency:
                        entry = {
                            vltk.qid: str(x["qa_id"]),
                            vltk.imgid: str(x["image_id"]),
                            vltk.text: x["question"],
                            vltk.label: [clean_label(x["answer"])],
                        }
                        batch_entries.append(entry)
                    else:
                        skipped += 1
        print(
            f"num skipped: {skipped / len(batch_entries)}, unfilterd num answers: {len(answer_counts)}"
        )
        return batch_entries

    # first entry in the dataset



import os

from transformers import BertTokenizerFast, RobertaTokenizerFast
from vltk import build
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig, VisionConfig

if __name__ == "__main__":
    datadir = os.path.join(os.environ["HOME"], "demodata")
    # Adapters().get("funsd").extract(datadir)
    config = DataConfig(
        # lang=LangConfig(tokenizer="BertWordPieceTokenizer"),
        train_datasets=[["funsd", "train"]],
        visn=VisionConfig(grayscale=True),
        lang=LangConfig(max_visual_seq_length=512, tokenizer=BertTokenizerFast),
        visn_processors=["auxtokenize", "ocrboxfixed", "tokenlabels", "xywhtoxyxy"],
        num_workers=1,
        extractor=None,
        datadir=datadir,
        train_batch_size=2,
        eval_batch_size=2,
        add_visual_cls=True,
        ignore_image=True,
        reextract=True,
        redownload=False,
    )

    train_loader, val_loader = build(config)
    for x in train_loader:
        pass



from collections import defaultdict

import numpy as np
import vltk.vars as vltk
from vltk import adapters
from vltk.features import Features


class CLEVR(adapters.VisnDataset):
    @staticmethod
    def schema(dim=3):
        return {
            "positions": Features.Features2D(d=dim),
            "colors": Features.StringList(),
            "shapes": Features.StringList(),
            "sizes": Features.StringList(),
            "materials": Features.StringList(),
        }

    @staticmethod
    def forward(json_files, splits):
        entries = defaultdict(dict)
        for filepath, js in json_files.items():
            for scene in js["scenes"]:
                img_filename = scene["image_filename"]
                imgid = img_filename.split(".")[0]
                objects = scene["objects"]
                colors, shapes, materials, sizes, positions = [], [], [], [], []
                for obj in objects:
                    colors.append(obj["color"])
                    shapes.append(obj["shape"])
                    materials.append(obj["material"])
                    sizes.append(obj["size"])
                    positions.append(obj["pixel_coords"])
                entries[imgid] = {
                    "positions": np.array(positions),
                    "colors": colors,
                    "shapes": shapes,
                    "materials": materials,
                    "sizes": sizes,
                    vltk.imgid: imgid,
                }

        return [v for v in entries.values()]



import os
from typing import Dict, List, Union

from PIL import Image as PImage

from vltk.abc import config
from vltk.inspection import get_args
from vltk.memory import get_most_free_gpu
# from vltk.loader.builder import init_datasets
from vltk.processing.image import Image

"""
Old config options that may be good to remember later:
pos_dim: int = 4
visual_dim: int = 2048
num_attrs: int = 400
num_objects: int = 1600
textfile_extensions: Union[List[str], str] = ["json", "jsonl"]
img_format: str = "jpg"

"""


class ModelConfig(config.Config):
    checkpoint = None
    freeze_layers = None
    freeze_embeddigs = None
    freeze_heads = None

    def __init__(self, **kwargs):
        for f, v in kwargs.items():
            setattr(self, f, v)
            self._overwritten[f] = v


class ModelsConfig(config.Config):
    main_model: str = "lxmert"
    checkpoint = None
    all_on_same_device = False
    models_to_devices = None

    def add(self, model_name, model_config):
        model_base = model_name.split("_")[0]
        attr_dict = {}
        if hasattr(self, model_base):
            for attr, attr_val in getattr(self, model_base).items():
                attr_dict[attr] = attr_val
            mconf = model_config(**attr_dict)

            setattr(self, model_base, mconf)
        else:
            raise Exception

    def __init__(self, **kwargs):
        for f, v in kwargs.items():

            if isinstance(v, dict):
                v = ModelsConfig(**v)
            setattr(self, f, v)
            self._overwritten[f] = v


class PretrainConfig(config.Config):
    epochs: int = 4
    task_matched: bool = False
    task_mask_lm: bool = False
    task_obj_predict: bool = False
    visual_attr_loss: bool = False
    visual_obj_loss: bool = False
    visual_feat_loss: bool = False


class EvalConfig(config.Config):
    half_precision: bool = True
    task_matched: bool = False
    task_mask_lm: bool = False
    task_obj_predict: bool = False
    visual_attr_loss: bool = False
    visual_obj_loss: bool = False
    visual_feat_loss: bool = False


class FinetuneConfig(config.Config):
    learning_rate: float = 1e-5
    half_precision: bool = True
    epochs: int = 4
    gamma: float = 0.01
    max_norm: float = 5.0
    warmup: float = 0.10
    weight_decay: float = 0.01
    task_matched: bool = False
    task_mask_lm: bool = False
    task_obj_predict: bool = False
    visual_attr_loss: bool = False
    visual_obj_loss: bool = False
    visual_feat_loss: bool = False


class LangConfig(config.Config):
    vocab_path_or_name: Union[None, str] = None
    tokenizer: Union[str, object] = "BertWordPieceTokenizer"
    word_mask_rate: float = 0.15  # these all belong to text preporcessors. change
    feature_mask_rate: float = 0.15
    random_feature_rate: float = 0.10
    random_word_rate: float = 0.10
    sentence_match_rate: float = 0.50
    truncate_sentence: bool = True
    return_token_type_ids: bool = True
    add_special_tokens: bool = True
    return_tensors: str = "pt"
    return_attention_mask: bool = True
    ignore_id: int = -100
    max_seq_length: int = 36
    max_visual_seq_length: int = 128
    lowercase: bool = False
    pad_direction: str = "right"


class VisionConfig(config.Config):
    transforms: List = ["FromFile", "Resize", "ToTensor"]
    interpolation = PImage.BICUBIC
    grayscale: bool = False
    size: tuple = (256, 256)

    def __init__(self, **kwargs):
        for f, v in kwargs.items():
            setattr(self, f, v)
            self._overwritten[f] = v

    def build(self):
        from torchvision.transforms import transforms

        _image = Image()
        kwargs = self.to_dict()
        transformlist = kwargs.pop("transforms")
        funcs = []
        for t in transformlist:
            name = t
            t = _image.get(t)
            if name == "Lambda":
                assert hasattr(
                    self, "lambd"
                ), "Lambda transform requires the keyword arg: lambd"
                funcs.append(t(self.lambd))
                continue
            args = get_args(t, kwargs)
            if args is None:
                funcs.append(t())
            else:
                funcs.append(t(**args))
        return transforms.Compose(funcs)


class DataConfig(config.Config):
    lang_processors: Union[None, List[str]] = None
    visn_processors: Union[None, List[str]] = None
    visnlang_processors: Union[None, List[str]] = None
    visn: Union[VisionConfig, dict] = {}
    lang: Union[LangConfig, dict] = {}
    labels: Union[None, str] = None
    eval_datasets = None
    train_datasets = None
    rand_feats: Union[None, tuple] = None  # if tuple, tuple represents shape
    eval_batch_size = 32
    train_batch_size = 64
    extractor: Union[None, str] = None
    datadir: str = None
    img_first: bool = False
    shuffle: bool = True
    num_workers: int = 8
    drop_last: bool = True
    pin_memory: bool = False
    percent: int = 1.0
    collate_simple: bool = True
    ignore_annotations: bool = False
    ignore_filepath: bool = True
    ignore_segmentation: bool = False
    ignore_image: bool = False
    metadata_filedict: Union[None, Dict[str, str]] = None
    add_visual_cls: bool = True
    reextract = False
    redownload = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        visn = kwargs.get("visn", {})
        lang = kwargs.get("lang", {})
        if isinstance(visn, VisionConfig):
            self.visn = visn
        else:
            self.visn = VisionConfig(**visn)
        if isinstance(lang, LangConfig):
            self.lang = lang
        else:
            self.lang = LangConfig(**lang)

    def build(self):
        raise NotImplementedError(
            "Do not import DataConfig from this file and try to build. Import Instead from `vltk.loader import\
            DataConfig"
        )
        pass


class Config(config.Config):
    data: DataConfig = None
    models: ModelsConfig = None
    eval: EvalConfig = None
    train: Union[FinetuneConfig, PretrainConfig] = None
    logging: bool = True
    gpu: int = None
    seed: int = 9595
    percent_min_gpu_free_mem: float = 0.75
    print_config: bool = False
    datasets: Union[None, str] = None
    base_logdir: str = os.path.join(os.environ.get("HOME", os.getcwd()), "logs")
    rel_logdir: str = ""
    logdir: str = (
        None  # this will be determined lated by datadir + base_logdir + rel_logdir
    )
    test_save: bool = False
    save_on_crash = False
    save_after_exp = True
    save_after_epoch = False
    email = None
    experimentdir: Union[None, str] = os.getcwd()
    test_run: bool = True
    break_loop_on_test: bool = True
    empty_cache: bool = True
    launch_blocking: bool = True
    vltk_checkpoint_dir: Union[str, None] = None

    def __init__(self, finetune=True, **kwargs):
        super().__init__(**kwargs)
        self.logdir = os.path.join(self.base_logdir, self.rel_logdir)
        if finetune:
            self.train = FinetuneConfig(**kwargs.get("train", {}))
        else:
            self.train = PretrainConfig(**kwargs.get("train", {}))

        self.eval = EvalConfig(**kwargs.get("eval", {}))
        self.data = DataConfig(**kwargs.get("data", {}))
        self.models = ModelsConfig(**kwargs.get("models", {}))

        self._set_gpus()

        if self.launch_blocking:
            os.environ["CUDA_LAUNCH_BLOCKING"] = "1"

        """
        ? a better way to propogate these to subconfigs ?
        """
        setattr(self.data, "test_run", self.test_run)
        setattr(self.data, "logdir", self.logdir)

    # should add more testcases incase all gpus are busy
    def _set_gpus(self):
        if self.gpu is not None:
            pass
        if self.gpu is None:
            self.gpu = get_most_free_gpu()

        if self.gpu == -1:
            print("WARNING: setting everything to cpu")
            self.gpu = "cpu"



# inspiration / code adapted from detectron2

import subprocess

import torch


# from facebook detectron
def handle_cuda_oom(func, *args, kwargs_1, kwargs_2):
    """
    A context which ignores CUDA OOM exception from pytorch.
    """
    try:
        yield func(*args, **kwargs_1)
    except RuntimeError as e:
        # NOTE: the string may change?
        if "CUDA out of memory. " in str(e):
            torch.cuda.empty_cache()
            yield func(*args, **kwargs_2)
        else:
            raise


def get_most_free_gpu():
    if not torch.cuda.is_available():
        return -1
    mem_list = get_nvidia_gpu_memory()
    return min(
        list(map(lambda k: (k, mem_list[k][0] / mem_list[k][1]), mem_list)),
        key=lambda x: x[1],
    )[0]


def get_nvidia_gpu_memory():
    result = subprocess.check_output(
        [
            "nvidia-smi",
            "--query-gpu=memory.used,memory.total",
            "--format=csv,nounits,noheader",
        ],
        encoding="utf-8",
    )
    gpu_memory = [eval(x) for x in result.strip().split("\n")]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map



"""
 coding=utf-8
 Copyright 2018, Antonio Mendoza Hao Tan, Mohit Bansal
 Adapted From Facebook Inc, Detectron2

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.import copy
 """
import sys
from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image

from .transformers_compat import img_tensorize


class ResizeShortestEdge:
    def __init__(self, short_edge_length, max_size=sys.maxsize):
        """
        Args:
            short_edge_length (list[min, max])
            max_size (int): maximum allowed longest edge length.
        """
        self.interp_method = "bilinear"
        self.max_size = max_size
        self.short_edge_length = short_edge_length

    def __call__(self, imgs):
        img_augs = []
        for img in imgs:
            h, w = img.shape[:2]
            # later: provide list and randomly choose index for resize
            size = np.random.randint(
                self.short_edge_length[0], self.short_edge_length[1] + 1
            )
            if size == 0:
                return img
            scale = size * 1.0 / min(h, w)
            if h < w:
                newh, neww = size, scale * w
            else:
                newh, neww = scale * h, size
            if max(newh, neww) > self.max_size:
                scale = self.max_size * 1.0 / max(newh, neww)
                newh = newh * scale
                neww = neww * scale
            neww = int(neww + 0.5)
            newh = int(newh + 0.5)

            if img.dtype == np.uint8:
                pil_image = Image.fromarray(img)
                pil_image = pil_image.resize((neww, newh), Image.BILINEAR)
                img = np.asarray(pil_image)
            else:
                img = img.permute(2, 0, 1).unsqueeze(0)  # 3, 0, 1)  # hw(c) -> nchw
                img = F.interpolate(
                    img, (newh, neww), mode=self.interp_method, align_corners=False
                ).squeeze(0)
            img_augs.append(img)

        return img_augs


class Preprocess:
    def __init__(self, cfg):
        self.aug = ResizeShortestEdge(
            [cfg.INPUT.MIN_SIZE_TEST, cfg.INPUT.MIN_SIZE_TEST], cfg.INPUT.MAX_SIZE_TEST
        )
        self.input_format = cfg.INPUT.FORMAT
        self.size_divisibility = cfg.SIZE_DIVISIBILITY
        self.pad_value = cfg.PAD_VALUE
        self.max_image_size = cfg.INPUT.MAX_SIZE_TEST
        self.device = cfg.MODEL.DEVICE
        self.pixel_std = (
            torch.tensor(cfg.MODEL.PIXEL_STD)
            .to(self.device)
            .view(len(cfg.MODEL.PIXEL_STD), 1, 1)
        )
        self.pixel_mean = (
            torch.tensor(cfg.MODEL.PIXEL_MEAN)
            .to(self.device)
            .view(len(cfg.MODEL.PIXEL_STD), 1, 1)
        )
        self.normalizer = lambda x: (x - self.pixel_mean) / self.pixel_std

    def pad(self, images):
        max_size = tuple(max(s) for s in zip(*[img.shape for img in images]))
        image_sizes = [im.shape[-2:] for im in images]
        images = [
            F.pad(
                im,
                [0, max_size[-1] - size[1], 0, max_size[-2] - size[0]],
                value=self.pad_value,
            )
            for size, im in zip(image_sizes, images)
        ]

        return torch.stack(images), torch.tensor(image_sizes)

    def __call__(self, images, img_ids):
        with torch.no_grad():
            tensor_imgs = []
            good_ids = []
            if not isinstance(images, list):
                assert isinstance(images, str)
                images = [images]
            for img_id, img in zip(img_ids, images):
                if isinstance(img, torch.Tensor):
                    tensor_imgs.append(img.to(self.device).float())
                    good_ids.append(img_id)
                elif not isinstance(img, torch.Tensor):
                    img = img_tensorize(img, input_format=self.input_format)
                    if img is not None:
                        tensor_imgs.append(torch.as_tensor(img).to(self.device).float())
                        good_ids.append(str(img_id))
                    else:
                        continue

            if len(tensor_imgs) == 0:
                return [], [], [], []

            # resize smallest edge
            images = tensor_imgs
            raw_sizes = torch.tensor([im.shape[:2] for im in images])
            images = self.aug(images)
            # transpose images and convert to torch tensors
            # images = [torch.as_tensor(i.astype("float32")).permute(2, 0, 1).to(self.device) for i in images]
            # now normalize before pad to aoid useless arithmatic
            images = [self.normalizer(x) for x in images]
            # now pad them to do the following operations
            images, sizes = self.pad(images)
            # Normalize

            if self.size_divisibility > 0:
                raise NotImplementedError()
            # pad
            scales_yx = torch.true_divide(raw_sizes, sizes)
            return good_ids, images, sizes, scales_yx


def _scale_box(boxes, scale_yx):
    boxes[:, 0::2] *= scale_yx[:, 1]
    boxes[:, 1::2] *= scale_yx[:, 0]
    return boxes


def _clip_box(tensor, box_size: Tuple[int, int]):
    assert torch.isfinite(tensor).all(), "Box tensor contains infinite or NaN!"
    h, w = box_size
    tensor[:, 0].clamp_(min=0, max=w)
    tensor[:, 1].clamp_(min=0, max=h)
    tensor[:, 2].clamp_(min=0, max=w)
    tensor[:, 3].clamp_(min=0, max=h)



from collections import Counter

from vltk.features import Features
from vltk import adapters


from vltk.features import Features
from vltk.utils.adapters import clean_label
import vltk.vars as vltk


class GQA(adapters.VisnLangDataset):
    data_info = {
        "dev": {"coco2014": ["test"]},
        "train": {"visualgenome": ["train"]},
        "val": {"visualgenome": ["train"]},
        "test": {"coco2014": ["test"]},
        "testdev": {"coco2014": ["val"]},
    }

    filters = ["unbalanced", "train"]

    @staticmethod
    def schema():
        return {vltk.label: Features.StringList(), "layout": Features.StringList()}

    @staticmethod
    def forward(json_files, split, min_label_frequency=2):
        skipped = 0
        label_frequencies = Counter()
        batch_entries = []

        for filename, data in json_files.items():
            for i, (k, v) in enumerate(data.items()):
                if "answer" in v:
                    answer = clean_label(v["answer"])
                    label_frequencies.update([answer])

            for i, (k, v) in enumerate(data.items()):
                if split == "test":
                    answer = None
                    layout = None
                elif label_frequencies[v["answer"]] < min_label_frequency:
                    skipped += 1
                    continue
                else:
                    answer = clean_label(v["answer"])
                    layout = [layout["operation"] for layout in v["semantic"]]

                text = v["question"]
                img_id = v["imageId"].lstrip("n")

                entry = {
                    vltk.text: text,
                    vltk.imgid: img_id,
                    vltk.label: [answer],
                    "layout": layout,
                }

                batch_entries.append(entry)

        return batch_entries



import os

from transformers import BertTokenizerFast, RobertaTokenizerFast
from vltk import build
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig, VisionConfig

if __name__ == "__main__":
    datadir = os.path.join(os.environ["HOME"], "data")
    Adapters().get("funsd").extract(datadir)
    config = DataConfig(
        # lang=LangConfig(tokenizer="BertWordPieceTokenizer"),
        train_datasets=[["gqa", "train"]],
        num_workers=1,
        extractor=None,
        datadir=datadir,
        train_batch_size=2,
        eval_batch_size=2,
        add_visual_cls=True,
        ignore_image=True,
    )

    train_loader, val_loader = build(config)
    for x in train_loader:
        pass



import os
import unittest

from vltk import get_data

TEST_PATH = os.path.dirname(os.path.realpath(__file__))


class TestGeneric(unittest.TestCase):

    # setup rando things like schema, etc
    # for tests, we will want to test each new method, plus a test extraction. The test extraction will have to be first
    # in order of most general to most specific tests

    def test_extraction_single_dir(self):
        # okay so the extraction single dir will
        pass

    def test_extaction_multi_dir(self):
        pass

    def add_text_dataset(self):
        pass

    def test_create_column_text(self):
        pass

    def test_append_column_text(self):
        pass

    def test_remove_column_text(self):
        pass

    def test_create_labeled_column_text(self):
        pass

    def test_append_labeled_column_text(self):
        pass

    def test_remove_labeled_column_text(self):
        pass


'''
useful methods:
    save_to_disk
    datasets.concatenate_datasets
    load from disk
'''



import timeit
from collections import defaultdict

import datasets as ds
import vltk.vars as vltk
from tqdm import tqdm
from vltk.adapters import VisnDataset
from vltk.processing.label import clean_imgid_default

# ignore 'iscrowd' because it used rle


class Coco2014(VisnDataset):
    @staticmethod
    def default_features(**kwargs):
        return {
            "bbox": ds.Sequence(
                length=-1, feature=ds.Sequence(length=-1, feature=ds.Value("float32"))
            ),
            "segmentation": ds.Sequence(
                length=-1,
                feature=ds.Sequence(
                    length=-1,
                    feature=ds.Sequence(length=-1, feature=ds.Value("float32")),
                ),
            ),
            "area": ds.Sequence(length=-1, feature=ds.Value("float32")),
            "size": ds.Sequence(length=-1, feature=ds.Value("int16")),
        }

    # add code to ensure correct format of all bounding boxes
    def forward(json_files):
        """ json_files: list(tuple(str, dict)) """

        total_annos = {}
        id_to_cat = {}
        id_to_size = {}
        for file, json in tqdm(json_files):
            if "instance" not in file:
                continue
            info = json["images"]
            for i in info:
                id_to_size[clean_imgid_default(i["file_name"]).split(".")[0]] = [
                    i["height"],
                    i["width"],
                ]

        for file, json in tqdm(json_files):
            if "instance" not in file:
                continue

            categories = json["categories"]
            for cat in categories:
                id_to_cat[cat["id"]] = cat["name"]

            for entry in tqdm(json["annotations"]):
                img_id = clean_imgid_default(str(entry["image_id"]))
                entry["size"] = id_to_size[img_id]
                bbox = entry["bbox"]
                area = entry["area"]
                segmentation = entry["segmentation"]
                category_id = id_to_cat[entry["category_id"]]
                if entry["iscrowd"]:
                    seg_mask = []
                else:
                    seg_mask = segmentation
                    if not isinstance(seg_mask[0], list):
                        seg_mask = [seg_mask]
                # seg_mask, size = compress_mask(
                #     seg_to_mask(segmentation, entry["size"][0], entry["size"][1])
                # )

                img_data = total_annos.get(img_id, None)
                if img_data is None:
                    img_entry = defaultdict(list)
                    img_entry["size"] = entry["size"]
                    img_entry[vltk.label].append(category_id)
                    img_entry["bbox"].append(bbox)
                    img_entry["area"].append(area)
                    img_entry["segmentation"].append(seg_mask)
                    total_annos[img_id] = img_entry
                else:
                    total_annos[img_id]["bbox"].append(bbox)
                    total_annos[img_id][vltk.label].append(category_id)
                    total_annos[img_id]["area"].append(area)
                    total_annos[img_id]["segmentation"].append(seg_mask)
                    # for sm in seg_mask:

        return [{vltk.imgid: img_id, **entry} for img_id, entry in total_annos.items()]


if __name__ == "__main__":

    coco = Coco2014.extract(
        searchdirs="/playpen1/home/avmendoz/data",
    )

    start = timeit.timeit()
    coco = Coco2014.load(
        "/playpen1/home/avmendoz/data/coco2014/annotations/annotations.arrow"
    )
    end = timeit.timeit()
    print(coco, f"duration to load: {end - start} s")

    print("example:", coco[0]["label"])



from itertools import chain

import torch
import vltk.vars as vltk
from vltk.processing import VisnProcessor
from vltk.utils.adapters import (imagepoints_to_mask, rescale_box,
                                 resize_binary_mask, seg_to_mask,
                                 truncate_and_pad_list)


class PolygonProcessor(VisnProcessor):
    def forward(self, entry, **kwargs):
        polykey = vltk.polygons
        size = entry[vltk.size]
        if vltk.rawsize not in entry:
            rawsize = size
        else:
            rawsize = entry[vltk.rawsize]
        segs = torch.stack(
            list(
                map(
                    lambda x: resize_binary_mask(seg_to_mask(x, *rawsize), size),
                    entry.pop(polykey),
                ),
            )
        )

        segs = segs[: min(len(segs), self.config.lang.max_visual_seq_length)]
        segs = torch.nn.functional.pad(
            segs,
            (0, 0, 0, 0, 0, self.config.lang.max_visual_seq_length - len(segs)),
        )
        entry[vltk.segmentation] = segs
        return entry


class RLEProcessor(VisnProcessor):
    def forward(self, entry, **kwargs):
        rlekey = vltk.RLE
        segs = torch.stack(
            list(
                map(
                    lambda x: resize_binary_mask(
                        imagepoints_to_mask(x, entry[vltk.rawsize]),
                        torch.as_tensor(entry[vltk.size]),
                    ),
                    entry.pop(rlekey),
                )
            )
        )
        segs = segs[: min(len(segs), self.config.lang.max_visual_seq_length)]
        segs = torch.nn.functional.pad(
            segs,
            (0, 0, 0, 0, 0, self.config.lang.max_visual_seq_length - len(segs)),
        )
        entry[vltk.segmentation] = segs


class AuxTokenize(VisnProcessor):
    _keys = vltk.text

    def enable_padding(self):
        self.tokenizer.enable_padding(
            length=self.config.lang.max_seq_length,
            direction=self.config.lang.pad_direction,
            pad_id=self.tokenizer.token_to_id(self.tokenizer.pad_token),
        )

    def disable_padding(self):
        self.tokenizer.no_padding()

    def forward(self, entry, **kwargs):
        max_len = self.config.lang.max_visual_seq_length
        text = entry.pop(vltk.text)
        if self.config.add_visual_cls:
            text = [self.tokenizer.cls_token] + text

        if not self.from_transformers:
            self.disable_padding()
            unk_id = self.tokenizer.token_to_id(self.tokenizer.unk_token)
            text = list(
                map(
                    lambda x: x.ids,
                    self.tokenizer.encode_batch(text, add_special_tokens=False),
                )
            )
            self.enable_padding()
        else:
            unk_id = self.tokenizer.convert_tokens_to_ids(self.tokenizer.unk_token)
            text = self.tokenizer(
                text,
                add_special_tokens=False,
                return_attention_mask=False,
            )["input_ids"]

        text = list(map(lambda x: x if x else [unk_id], text))

        tokenmap = list(map(lambda x: len(x), text))
        if len(tokenmap) >= max_len:
            tokenmap = tokenmap[: max_len - 1]

        assert 0 not in tokenmap
        tokenmap = torch.tensor(
            truncate_and_pad_list(tokenmap, max_len, self.config.lang.ignore_id)
        )
        entry[vltk.tokenmap] = tokenmap
        text = list(chain(*text))
        visual_attention_mask = torch.tensor(
            [1] * min(max_len, len(text)) + [0] * max(0, max_len - len(text))
        )
        entry["visual_attention_mask"] = visual_attention_mask
        if not self.from_transformers:
            pad_id = self.tokenizer.token_to_id(self.tokenizer.pad_token)
            text = truncate_and_pad_list(text, max_len - 1, pad_id)
            text += [self.tokenizer.token_to_id(self.tokenizer.sep_token)]
        else:
            pad_id = self.tokenizer.convert_tokens_to_ids(self.tokenizer.pad_token)
            text = truncate_and_pad_list(text, max_len - 1, pad_id)
            text += [self.tokenizer.convert_tokens_to_ids(self.tokenizer.sep_token)]

        entry[vltk.text] = torch.tensor(text)
        return entry


class OCRBox(VisnProcessor):
    _keys = (vltk.tokenbox, vltk.tokenmap)

    def forward(self, entry, **kwargs):
        max_len = self.config.lang.max_visual_seq_length
        tokenboxes = entry.pop(vltk.tokenbox)
        if self.config.add_visual_cls:
            tokenboxes = [[0, 0, *entry[vltk.rawsize]]] + tokenboxes
        tokenmap = entry.get(vltk.tokenmap)
        tokenboxes = list(
            chain(
                *map(
                    lambda x: [x[0]] * x[1],
                    zip(tokenboxes, tokenmap),
                )
            )
        )
        tokenboxes = truncate_and_pad_list(tokenboxes, max_len, [0, 0, 0, 0])
        tokenboxes = torch.tensor(tokenboxes)
        if vltk.size in entry:
            tokenboxes = rescale_box(tokenboxes, entry[vltk.scale])
        entry[vltk.tokenbox] = tokenboxes
        return entry


class TokenLabels(VisnProcessor):
    _keys = (vltk.label, vltk.tokenmap)

    def forward(self, entry, **kwargs):
        max_len = self.config.lang.max_visual_seq_length
        labels = entry.get(vltk.label)
        if self.config.add_visual_cls:
            labels = [""] + labels
        tokenmap = entry.get(vltk.tokenmap)
        labels = list(
            chain(
                *map(
                    lambda x: [x[0]] * x[1],
                    zip(labels, tokenmap),
                )
            )
        )
        if len(labels) >= max_len:
            labels = labels[: max_len - 1]
        entry[vltk.label] = labels
        return entry


class OCRBoxFixed(VisnProcessor):
    _keys = (vltk.tokenbox, vltk.tokenmap, vltk.rawsize)

    def forward(self, entry, **kwargs):
        max_len = self.config.lang.max_visual_seq_length
        tokenboxes = entry.pop(vltk.tokenbox)
        raw_w, raw_h = entry[vltk.rawsize]
        scale = (1000 / raw_w, 1000 / raw_h)
        if self.config.add_visual_cls:

            tokenboxes = [[0, 0, raw_w, raw_h]] + tokenboxes
        tokenmap = entry.get(vltk.tokenmap)
        tokenboxes = list(
            chain(
                *map(
                    lambda x: [x[0]] * x[1],
                    zip(tokenboxes, tokenmap),
                )
            )
        )
        tokenboxes = truncate_and_pad_list(tokenboxes, max_len, [0, 0, 0, 0])
        tokenboxes = torch.tensor(tokenboxes)
        tokenboxes = torch.clamp(rescale_box(tokenboxes, scale), min=0, max=1000)
        entry[vltk.tokenbox] = tokenboxes
        return entry


class XYWHtoXYXY(VisnProcessor):
    def forward(self, entry, **kwargs):
        for k in (vltk.tokenbox, vltk.box):
            if k in entry:
                box = entry[k]
                try:
                    box[:, -2:] += box[:, :2]
                except Exception:
                    raise Exception(box)
                entry[k] = box
        return entry



from transformers import RobertaTokenizerFast
from vltk import build
from vltk.adapters import Adapters
from vltk.configs import DataConfig, LangConfig
from vltk.processing import LangProcessor


class TestProcessor(LangProcessor):
    def forward(self, x, *args, **kwargs):
        return x


if __name__ == "__main__":
    get_loaders_visnlang = True
    get_loaders_visn = False
    extract_visn = False
    extract_visnlang = False
    extract_extractor = False
    datadir = "/home/eltoto/demodata"
    use_extractor = False
    if extract_visnlang:
        Adapter = Adapters().get("vqa")
        result = Adapter.extract(datadir)["val"]
        result = Adapter.load(datadir)["val"]
    if extract_visn:
        Adapter = Adapters().get("coco2014")
        result = Adapter.extract(datadir)
        result = Adapter.load(datadir)
    if extract_extractor:
        Adapter = Adapters().get("frcnn")
        result = Adapter.extract(datadir, dataset="coco2014")
        result = Adapter.load(datadir)
    if get_loaders_visnlang:
        # add adapters to library

        # config
        config = DataConfig(
            # choose which dataset and dataset split for train and eval
            train_datasets=[
                ["gqa", "train"],
                ["vqa", "trainval"],
                ["cococaptions", "trainval"],
                ["vgqa", "train"],
            ],
            # choose which feature extractor to use
            extractor="frcnn",
            datadir=datadir,
            num_workers=8,
            train_batch_size=1,
            # iterate with through datasets via images first versus text
            img_first=True,
            # ignore segmentation annotations from being prcoessed with the COCO dataset
            ignore_segmentation=True,
        )
        train, val = init_datasets(config)
        for b in train:
            print(b.keys())
            pass

    if get_loaders_visn:

        config = DataConfig(
            train_datasets=[
                ["coco2014", "train"],
            ],
            datadir=datadir,
            num_workers=1,
            train_batch_size=1,
            eval_batch_size=1,
        )

        train_loader, val_loader = build(config)
        for x in train_loader:
            print(x)
            break



from collections import Counter

import datasets as ds
from tqdm import tqdm
from vltk.abc.visnlangdatasetadapter import VisnLangDatasetAdapter
from vltk.metrics import soft_score

TESTPATH = "/playpen1/home/avmendoz/data/vqa/train"


# user must only define forward in this function, and dataset features


class VQAset(VisnLangDatasetAdapter):
    name = "vqa"
    data_info = {
        "val": {"coco2014": ["val"]},
        "train": {"coco2014": ["train"]},
        "test": {"coco2014": ["test"]},
    }
    default_features = {
        "qid": ds.Value("string"),
    }

    def forward(text_data, split, label_preprocessor=None, **kwargs):
        min_label_frequency = kwargs.get("min_label_frequency")
        batch_entries = []
        all_questions = []
        qid2answers = {}
        label_frequencies = Counter()
        if label_preprocessor is None:

            def label_preprocessor(x):
                return x

        for x in tqdm(text_data):
            if "questions" in x:
                all_questions.extend(x["questions"])
            else:
                annotations = x["annotations"]
                accepted_answers = {
                    label_preprocessor(anno["multiple_choice_answer"])
                    for anno in annotations
                }
                for anno in tqdm(annotations):
                    qid = str(anno["question_id"])
                    answers = anno["answers"]
                    label_frequencies.update(
                        [label_preprocessor(anno["multiple_choice_answer"])]
                    )
                    answer_counter = Counter()
                    for ans_dict in answers:
                        ans = ans_dict["answer"]
                        if ans not in accepted_answers:
                            pass
                        else:
                            ans = label_preprocessor(ans)
                            # make  sure to clean label before updating frequncies
                            # label_frequencies.update([ans])
                            answer_counter.update([ans])
                    qid2answers[qid] = {
                        k: soft_score(v) for k, v in answer_counter.items()
                    }

        skipped = 0
        for entry in tqdm(all_questions):
            entry[VisnLangDatasetAdapter.img_key] = str(entry.pop("image_id"))
            entry[VisnLangDatasetAdapter.text_key] = entry.pop("question")
            entry["qid"] = str(entry.pop("question_id"))
            try:
                entry[VisnLangDatasetAdapter.label_key] = qid2answers[entry["qid"]]
                labels = {
                    l: s
                    for l, s in entry[VisnLangDatasetAdapter.label_key].items()
                    if label_frequencies[l] > min_label_frequency
                }
                if not labels:
                    skipped += 1
                    continue

                labels, scores = VisnLangDatasetAdapter._label_handler(labels)
                entry[VisnLangDatasetAdapter.score_key] = scores
                entry[VisnLangDatasetAdapter.label_key] = labels
            except KeyError:
                pass

            batch_entries.append(entry)

        print(f"SKIPPEd {skipped} entries")
        return batch_entries


if __name__ == "__main__":

    from vltk.configs import Config

    config = Config().data
    config.min_label_frequency = 9
    VQAset.extract(
        config=config,
        splits=["train", "val", "test"],
    )
    val = VQAset.from_config(config, splits="val")["val"]
    # get min frequency of answers when loading, so we know the lenngth right away

    # get path to arrow file
    # val.get_arrow_split(datadirs=config.datadirs, extractor="frcnn", split="val")
    # get path to raw img ids
    print(len(val.labels))

    # print("entry at row 1:", val.get_row(1))
    # print("entries with img id 262148:", val.get_from_img("262148"))
    # print("freq of answer table:", val.get_freq("table"))
    # print("num_labels", val.num_labels)



import inspect
import os
import resource
import sys
from abc import ABCMeta
from copy import deepcopy

import torch
import vltk.vars as vltk
from datasets import Dataset
# disable logging from datasets
from datasets.utils.logging import set_verbosity_error
from vltk.utils.adapters import truncate_and_pad_list
from vltk.utils.base import convertids_recursive

# note if we do not immport a pacakage correctly in this class, no loops or exps will be present


__import__("tokenizers")
TOKENIZERS = {
    m[0]: m[1] for m in inspect.getmembers(sys.modules["tokenizers"], inspect.isclass)
}

rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (6144, rlimit[1]))

set_verbosity_error()


os.environ["TOKENIZERS_PARALLELISM"] = "False"


class SplitRangesVision:
    def __init__(self, nested_dict):
        # takes nested dict where the top level is the dataset name
        # the lower level keys are the split names

        self.nested_dict = nested_dict
        self.range2dataset = {}
        self.subrange2dataset = {}
        self.range2split = {}
        uniq_imgs = ()
        start = 0
        dataset_len = 0

        for dataset_name, dataset_dict in nested_dict.items():
            cur_dataset_len = 0
            for split_name, data in dataset_dict.items():
                rng = range(start, len(data) + start)
                self.subrange2dataset[rng] = dataset_name
                self.range2split[rng] = split_name
                start += len(data)
                cur_dataset_len += len(data)

                try:
                    img_ids = data.imgids
                except Exception:
                    img_ids = data.keys()

                img_ids = tuple(img_ids)

                # check once

                uniq_imgs += img_ids

            self.range2dataset[
                range(dataset_len, dataset_len + cur_dataset_len)
            ] = dataset_name
            dataset_len += cur_dataset_len

        self._len = dataset_len
        self.imgs = uniq_imgs
        self.uniq_imgs = uniq_imgs

    def __len__(self):
        return len(self.imgs)

    def __getitem__(self, x):
        if x >= len(self):
            raise IndexError(f"index {x} is out of range 0 to {len(self)}")
        for rng in self.range2split:
            if x in rng:
                visn_split = self.range2split[rng]
                visn_name = self.subrange2dataset[rng]
                return visn_name, visn_split


class SplitRangesVL:
    def __init__(self, nested_dict):
        # takes nested dict where the top level is the dataset name
        # the lower level keys are the split names

        self.nested_dict = nested_dict
        self.range2dataset = {}
        self.subrange2dataset = {}
        self.range2split = {}
        self.uniq_imgs = set()
        start = 0
        dataset_len = 0
        for dataset_name, dataset_dict in nested_dict.items():
            cur_dataset_len = 0
            for split_name, data in dataset_dict.items():
                rng = range(start, len(data) + start)
                self.uniq_imgs.update(set(data.imgids))
                self.subrange2dataset[rng] = dataset_name
                self.range2split[rng] = split_name
                start += len(data)
                cur_dataset_len += len(data)

            self.range2dataset[
                range(dataset_len, dataset_len + cur_dataset_len)
            ] = dataset_name
            dataset_len += cur_dataset_len

        self._len = dataset_len
        self.uniq_imgs = tuple(self.uniq_imgs)

    def __len__(self):
        return self._len

    def __getitem__(self, x):
        if x >= len(self):
            raise IndexError(f"index {x} is out of range 0 to {len(self)}")
        for rng in self.range2split:
            if x in rng:
                lang_split = self.range2split[rng]
                langset_name = self.subrange2dataset[rng]
                # note does not support matching mulitple vision datasets to a single split
                langset = self.nested_dict[langset_name][lang_split]
                vinset_info = langset.data_info[lang_split]
                visnset_name = next(iter(vinset_info.keys()))
                vinset_splits = tuple(vinset_info.values())[0]
                return (lang_split, langset_name, visnset_name, vinset_splits)


class CollatedVLSets:
    def __init__(self, *args):

        self.args = args
        self.range2listpos = {}
        self.uniq_imgs = set()
        start = 0
        for i, a in enumerate(args):
            self.uniq_imgs.update(a.imgids)
            self.range2listpos[range(start, len(a) + start)] = i
            start += len(a)

    # TODO: better solution for more datasets
    def get(self, img_id, return_dataset=False):
        for adapter in self.args:
            # print(adapter, img_id)
            try:
                return adapter.get(img_id)
            except KeyError:
                pass
        raise Exception("image id not found in any  visndatasetadapter annotations")

    def __getitem__(self, x):
        if x >= len(self):
            raise IndexError(f"index {x} is out of range 0 to {len(self)}")
        for rng in self.range2listpos:
            if x in rng:
                listpos = self.range2listpos[rng]
                listind = x - rng.start
                return self.args[listpos][listind]

    def __len__(self):
        return sum(map(lambda x: len(x), self.args))

    def __iter__(self):
        return iter(map(lambda x: self[x], range(0, len(self))))


class CollatedVisionSets:
    def __init__(self, *args):

        self.args = args
        self.range2listpos = {}
        self.imgids = ()
        start = 0
        for i, a in enumerate(args):
            if isinstance(a, ABCMeta):
                continue
            self.range2listpos[range(start, len(a) + start)] = i
            start += len(a)

    # TODO: better solution for more datasets
    def get(self, img_id):
        for adapter in self.args:
            # if not isinstance(adapter, Dataset):
            try:
                return adapter.get(img_id)
            except Exception:
                continue
        return {}

        # raise Exception("image id not found in any  visndatasetadapter annotations")

    def __getitem__(self, x):
        if x >= len(self):
            raise IndexError(f"index {x} is out of range 0 to {len(self)}")
        for rng in self.range2listpos:
            if x in rng:
                listpos = self.range2listpos[rng]
                listind = x - rng.start
                return (
                    self.args[listpos][listind],
                    type(self.args[listpos]).__name__.lower(),
                )

    def __len__(self):
        return sum(map(lambda x: len(x), self.args))

    def __iter__(self):
        return iter(map(lambda x: self[x], range(0, len(self))))


class BaseDataset(Dataset):
    def _init_tokenizer(self, config):
        from_transformers = True
        if isinstance(config.tokenizer, str):
            try:
                self.tokenizer = TOKENIZERS[config.tokenizer](
                    vltk.VOCABPATH
                    if config.vocab_path_or_name is None
                    else config.vocab_path_or_name,
                    lowercase=config.lowercase,
                )
                try:
                    self.special_tokens = set(
                        [
                            self.tokenizer.bos_token,
                            self.tokenizer.eos_token,
                            self.tokenizer.unk_token,
                            self.tokenizer.sep_token,
                            self.tokenizer.pad_token,
                            self.tokenizer.cls_token,
                            self.tokenizer.mask_token,
                        ]
                    )
                except Exception:
                    self.tokenizer.bos_token = "[BOS]"
                    self.tokenizer.eos_token = "[EOS]"
                    self.tokenizer.unk_token = "[UNK]"
                    self.tokenizer.sep_token = "[SEP]"
                    self.tokenizer.pad_token = "[PAD]"
                    self.tokenizer.cls_token = "[CLS]"
                    self.tokenizer.mask_token = "[MASK]"
                    self.special_tokens = set(
                        [
                            self.tokenizer.bos_token,
                            self.tokenizer.eos_token,
                            self.tokenizer.unk_token,
                            self.tokenizer.sep_token,
                            self.tokenizer.pad_token,
                            self.tokenizer.cls_token,
                            self.tokenizer.mask_token,
                        ]
                    )

                # self.tokenizer.add_special_tokens(self.special_tokens)
                self.tokenizer.enable_truncation(max_length=config.max_seq_length)
                self.tokenizer.enable_padding(length=config.max_seq_length)
                special_ids = set(
                    [self.tokenizer.token_to_id(t) for t in self.special_tokens]
                )
                self.tokenizer.eos_id = self.tokenizer.token_to_id(
                    self.tokenizer.eos_token
                )
                self.tokenizer.unk_id = self.tokenizer.token_to_id(
                    self.tokenizer.unk_token
                )
                self.tokenizer.sep_id = self.tokenizer.token_to_id(
                    self.tokenizer.sep_token
                )
                self.tokenizer.pad_id = self.tokenizer.token_to_id(
                    self.tokenizer.pad_token
                )
                self.tokenizer.cls_id = self.tokenizer.token_to_id(
                    self.tokenizer.cls_token
                )
                self.tokenizer.mask_id = self.tokenizer.token_to_id(
                    self.tokenizer.mask_token
                )
                self.special_ids = deepcopy(special_ids)
            except KeyError:
                raise Exception(
                    f"{config.tokenizer} not available. Try one of: {TOKENIZERS.keys()}.\
                            OR pass a Tokenizer class from transformers"
                )
            from_transformers = False
        else:
            # raise Exception("/".join(VOCABPATH.split("/")))
            if config.vocab_path_or_name is None:

                self.tokenizer = config.tokenizer.from_pretrained(
                    pretrained_model_name_or_path="/".join(
                        vltk.VOCABPATH.split("/")[:-1]
                    )
                )
            else:
                self.tokenizer = config.tokenizer.from_pretrained(
                    config.vocab_path_or_name
                )
            self.special_tokens = set(
                [
                    self.tokenizer.bos_token,
                    self.tokenizer.eos_token,
                    self.tokenizer.unk_token,
                    self.tokenizer.sep_token,
                    self.tokenizer.pad_token,
                    self.tokenizer.cls_token,
                    self.tokenizer.mask_token,
                ]
            )
            special_ids = set(
                [self.tokenizer.convert_tokens_to_ids(t) for t in self.special_tokens]
            )
            self.special_ids = deepcopy(special_ids)
            self.tokenizer.eos_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.eos_token
            )
            self.tokenizer.unk_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.unk_token
            )
            self.tokenizer.sep_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.sep_token
            )
            self.tokenizer.pad_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.pad_token
            )
            self.tokenizer.cls_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.cls_token
            )
            self.tokenizer.mask_id = self.tokenizer.convert_tokens_to_ids(
                self.tokenizer.mask_token
            )
        self.from_transformers = from_transformers

        all_ids = tuple(i[1] for i in self.tokenizer.get_vocab().items())
        self.all_ids = all_ids

    @property
    def batch_size(self):
        if not self.is_train:
            return self.config.eval_batch_size
        else:
            return self.config.train_batch_size

    def try_tensorify(self, entry):
        for k in entry:
            if not isinstance(entry[k], torch.Tensor):
                try:
                    entry[k] = torch.tensor(entry[k])
                except Exception:
                    pass
            if not isinstance(entry[k], torch.Tensor) and k in self.metadata_ids:
                meta = entry[k]
                if isinstance(entry[k][0], str) and isinstance(entry[k], list):
                    max_len = self.config.lang.max_visual_seq_length
                    meta = truncate_and_pad_list(meta, max_len, "")
                entry[k] = convertids_recursive(meta, self.metadata_ids[k])

        return entry

    def _update_placeholders(self, entry):
        raise NotImplementedError
        # if self.all_same_keys:
        #     return
        # entry_keys =
        # placeholders = self.placeholders
        # entry_keys



from vltk import OPTIMPATH
from vltk.inspection import import_funcs_from_file


class Optim:
    def __init__(self):
        if "OPTIMDICT" not in globals():
            global OPTIMDICT
            OPTIMDICT = import_funcs_from_file(OPTIMPATH, pkg="vltk.processing")

    def avail(self):
        return list(OPTIMDICT.keys())

    def get(self, name):
        return OPTIMDICT[name]

    def add(self, name, lab):
        OPTIMDICT[name] = lab



from collections import Counter

import vltk.vars as vltk
from vltk import adapters
from vltk.features import Features
from vltk.utils.adapters import clean_label, soft_score


# Vision-Language Datasets
class VQA(adapters.VisnLangDataset):
    data_info = {
        "val": {"coco2014": ["val"]},
        "train": {"coco2014": ["train"]},
        "test": {"coco2014": ["test"]},
    }

    @staticmethod
    def schema():
        # img id and score are assumed to be default features
        return {
            vltk.qid: Features.String(),
            vltk.label: Features.StringList(),
            vltk.score: Features.FloatList(),
        }

    @staticmethod
    def adjust_imgid(imgid, vdset_name, vdset_split):
        # length of COCO ids are are always length 12
        imgid = f'{"COCO"}_{vdset_split[0].lower()}{2014}_{"".join(["0"] * (12 - len(imgid)))}{imgid}'
        return imgid

    @staticmethod
    def forward(json_files, split, min_label_frequency=9):
        batch_entries = []
        all_questions = []
        qid2answers = {}
        label_frequencies = Counter()
        for filename, x in json_files.items():
            if "questions" in x:
                all_questions.extend(x["questions"])
            else:
                annotations = x["annotations"]
                accepted_answers = {
                    clean_label(anno["multiple_choice_answer"]) for anno in annotations
                }
                for anno in annotations:
                    qid = str(anno["question_id"])
                    answers = anno["answers"]
                    label_frequencies.update(
                        [clean_label(anno["multiple_choice_answer"])]
                    )
                    answer_counter = Counter()
                    for ans_dict in answers:
                        ans = ans_dict["answer"]
                        if ans not in accepted_answers:
                            pass
                        else:
                            ans = clean_label(ans)
                            answer_counter.update([ans])
                    qid2answers[qid] = {
                        k: soft_score(v) for k, v in answer_counter.items()
                    }

        skipped = 0
        for entry in all_questions:
            try:
                entry[vltk.imgid] = str(entry.pop("image_id"))
            except Exception:
                raise Exception(entry.keys())
            entry[vltk.text] = entry.pop("question")
            # entry.pop("question_id")
            entry["qid"] = str(entry.pop("question_id"))
            try:
                entry[vltk.label] = qid2answers[entry["qid"]]
                labels = {
                    l: s
                    for l, s in entry[vltk.label].items()
                    if label_frequencies[l] > min_label_frequency
                }
                if not labels:
                    skipped += 1
                    continue

                labels, scores = adapters.VisnLangDataset._label_handler(labels)
                entry[vltk.score] = scores
                entry[vltk.label] = labels
            except KeyError:
                pass

            batch_entries.append(entry)
        return batch_entries



# we can excpet list of forward fucntions at some other point
# for now, all we really need to care about is one forward method
# and how to convert that to eval if no class

import datetime
import gc
import json
import os
import random
import sys
from abc import ABC, abstractmethod
from collections import Iterable, OrderedDict, defaultdict
from copy import deepcopy
from statistics import mean
from typing import Dict, List, Union

import torch
import torch.nn as nn
from tqdm import tqdm
from transformers import AdamW, get_linear_schedule_with_warmup
from vltk import SIMPLEPATH, 
from vltk.abc.visndatasetadapter import VisnDatasetAdapters
from vltk.abc.visnlangdatasetadapter import VisnLangDatasetAdapters
# from vltk.dataset import UniversalLoader
from vltk.inspect import get_classes
from vltk.loader.builder import init_datasets
from vltk.modeling import Get as Mget
from vltk.modeling.configs import Get

__all__ = ["SimpleExperiment", "SimpleIdentifier", "SimpleExperiments"]
_visnlangdatasetadapters = VisnLangDatasetAdapters()
_visndatasetadapters = VisnDatasetAdapters()

# find all packages and classes when pointed to a specific directory?


class SimpleExperiments:
    def __init__(self):
        if "SIMPLEDICT" not in globals():
            global SIMPLEDICT
            SIMPLEDICT = get_classes(SIMPLEPATH, SimpleIdentifier, pkg="vltk.simple")

    def avail(self):
        return list(SIMPLEDICT.keys())

    def get(self, name):
        return SIMPLEDICT[name]

    def add(self, experiment):
        name = experiment.__name__
        SIMPLEDICT[name] = experiment


class SimpleIdentifier:
    pass


class SimpleExperiment(SimpleIdentifier, ABC):
    cur_epoch: int = 1
    cur_step: int = 1

    def __init__(self, config, datasets):
        self.config = config
        self.datasets = datasets
        self.__currently_training = False
        self.epochs = self.config.train.epochs
        self.cur_epoch = 1
        self._init_dirs()
        self._init_seed()
        self._init_scaler()
        # self._init_datasets()
        self._init_loaders()
        self._init_models()
        self._init_optim()
        self._init_checkpoint()

        # do we benchmark?
        if not self.config.empty_cache:
            torch.backends.benchmark = True

    def garbage_collect(self):
        if self.config.empty_cache:
            gc.collect()
            # torch.cuda.empty_cache()

    def currently_training(self):
        return True if self.__currently_training == "train" else False

    def _init_checkpoint(self):
        checkpoint_dir = self.config.vltk_checkpoint_dir
        if checkpoint_dir is None or not checkpoint_dir:
            return
        highest_model_epoch = {}
        # we can think about making sure the label size is right later
        for f in os.listdir(checkpoint_dir):
            path = os.path.join(checkpoint_dir, f)
            if "info" in path:
                exp_outputs = json.load(open(path))
                if "scheduler" in exp_outputs:
                    self.scheduler.load_state_dict(exp_outputs["scheduler"])
                self.cur_epoch = exp_outputs["cur_epoch"]
                self.epochs = exp_outputs["epochs"]
                self.cur_step = exp_outputs["cur_steps"]
            if "_epoch_" in path:
                if len(tuple(f.split(".")[0].split("_"))) == 3:
                    model_n, _, epoch_n = tuple(f.split(".")[0].split("_"))

                    highest_model_epoch[model_n] = max(
                        highest_model_epoch.get(model_n, 0), int(epoch_n)
                    )

        # load the model weights
        for model_n, highest_epoch in highest_model_epoch.items():
            f = os.path.join(checkpoint_dir, f"{model_n}_epoch_{highest_epoch}.pt")
            model = getattr(self, model_n)
            print(f"reloading weights for {model_n} for epoch {highest_epoch}")

            checkpoint = torch.load(f)
            keys = deepcopy(list(checkpoint.keys()))
            for k in keys:
                k_stripped = k.replace("module.", "")
                checkpoint[k_stripped] = checkpoint.pop(k)
            model.load_state_dict(checkpoint)
            # setattr(self, model_n, model)

    def _init_scaler(self):
        # all devices must be on some gpu if we want to use scaler

        half_precision = getattr(self.config.train, "half_precision", False)
        self.half_precision = half_precision
        if self.config.gpu == "cpu":
            half_precision = False
        self.scaler = None if not half_precision else torch.cuda.amp.GradScaler()

    def _init_models(self):
        model_list = self.model_list
        model_dict = {}
        model_configs = {}
        state_dict = None
        default_device = self.config.gpu if torch.cuda.is_available() else -1

        for x in model_list:
            if isinstance(x, tuple):
                name = x[0]
                model_class = x[1]
            elif isinstance(x, str):
                name = x
                model_class = Mget[name]

            model_config = getattr(self.config.models, name, None)
            model_device = getattr(model_config, "device", default_device)
            setattr(self, f"name_{model_device}", model_device)
            # TODO:  add a preinitialized option

            checkpoint = getattr(model_config, "checkpoint", None)
            print(f"instantiating {name} from {checkpoint}")

            # CASE 1
            if (
                checkpoint is not None
                and hasattr(model_class, "from_pretrained")
                and not os.path.isfile(checkpoint)
            ):
                # this is a huggingface model, so the config must be added appropriately
                model_instance = model_class.from_pretrained(
                    checkpoint,
                )
                checkpoint = model_instance.state_dict()
                model_config = Get[name](**model_config.to_dict())
                model_instance = model_class(model_config)

            # CASE 2
            elif not hasattr(model_class, "from_pretrained"):
                model_instance = model_class(**model_config.to_dict())
                if checkpoint is not None:
                    state_dict = torch.load(checkpoint)
            # CASE 3
            else:
                try:
                    model_instance = model_class(model_config)
                except Exception:
                    model_instance = model_class(**model_config.to_dict())

            # CASE 4: try loading checkpoint one last time
            if checkpoint is not None and state_dict is not None:

                # 1. filter out unnecessary keys
                new_state_dict = OrderedDict(
                    {
                        k: v
                        for k, v in state_dict.items()
                        if k in model_instance.state_dict()
                        and v.shape
                        == model_instance.state_dict()
                        .get(k, torch.tensor([], requires_grad=False))
                        .shape
                    }
                )
                # 3. load the new state dict
                model_instance.load_state_dict(new_state_dict, strict=False)

            # for question answering models, we will need to resize the number of labels
            # accordingly
            if hasattr(model_instance, "resize_num_qa_labels"):
                assert (
                    getattr(self, "label_to_id", None) is not None
                ), "no label dict found"
                print(f"Number of Labels: {len(self.label_to_id)}")
                label_file = self.config.data.labels
                if label_file is not None or "":
                    model_instance.resize_num_qa_labels(
                        len(json.load(open(label_file)))
                    )
                else:
                    model_instance.resize_num_qa_labels(len(self.label_to_id))

            # SET DEVICE
            if isinstance(model_device, int):
                model_instance.to(torch.device(model_device))
            elif isinstance(model_device, list) and len(model_device) > 1:
                os.environ["CUDA_LAUNCH_BLOCKING"] = "0"
                model_instance.to(torch.device(model_device[0]))
                model_instance = nn.DataParallel(
                    model_instance, device_ids=model_device
                ).to(torch.device(model_device[0]))
            elif isinstance(model_device, list) and len(model_device) == 1:
                model_instance.to(torch.device(model_device[0]))
            setattr(self, f"{name}_dev", model_device)

            model_dict[name] = model_instance
            model_configs[name] = model_config
            self.init_grad(name, model_instance)
            setattr(self, name, model_instance)

        self._model_dict = model_dict
        self._model_configs = model_configs

    # def _init_loader(self, visnlangdatasetadapterdict, visndatasetadapterdict, label_dict, train=True):
    #     datasets = self.datasets if train else self.config.data.eval_datasets
    #     loader = UniversalLoader(
    #         config=self.config.data,
    #         names=datasets,
    #         label_dict=label_dict,
    #         visndatasetadapterdict=visndatasetadapterdict,
    #         visnlangdatasetadapterdict=visnlangdatasetadapterdict,
    #     )
    #     # get necessary methods from the loader
    #     self.flatten_text = loader.dataset.flatten_text
    #     return loader

    @property
    def is_train(self):
        return getattr(self, "any_train", None)

    def _init_loaders(self):
        loaders, any_train, answer_to_id, object_to_id = init_datasets(self.config.data)
        self._loaders = loaders
        self.answer_to_id = answer_to_id
        self.object_to_id = object_to_id
        self.any_train = any_train
        # ttsd = self.train_visnlangdatasetadapterdict
        # tisd = self.train_visndatasetadapterdict
        # etsd = self.eval_visnlangdatasetadapterdict
        # eisd = self.eval_visndatasetadapterdict
        # l2id = self.label_to_id

        # loaders = {
        #     "train": self._init_loader(ttsd, tisd, l2id) if ttsd else None,
        #     "eval": self._init_loader(etsd, eisd, l2id) if etsd else None,
        # }
        # self._loaders = [(k, v) for k, v in loaders.items()]
        # self.is_train = any(map(lambda x: x == "train", [k for k in loaders]))
        # self._loaders = sorted(self._loaders, key=lambda x: x[0], reverse=True)
        # for k, v in loaders.items():
        #     if v is not None:
        #         self.transpose_img2txt = v.dataset.transpose_img2txt

    def _clean_dict(self, info):
        clean_info = {}
        for k, v in info.items():
            if (
                v is None
                or isinstance(v, torch.Tensor)
                or (
                    isinstance(v, Iterable)
                    and len(v) > 1
                    and isinstance(v[0], torch.Tensor)
                )
                or (not isinstance(v, bool) and not v)
            ):
                continue
            else:
                clean_info[k] = v

        return clean_info

    # other methods, too many methods
    def _init_optim(self):
        if self.is_train and self.model_dict:
            parameters = []
            for k, v in self.model_dict.items():
                parameters.extend(v.parameters())
            assert parameters, "no parameters added to optimizer"
            self._optim = AdamW(
                parameters,
                lr=self.config.train.learning_rate,
                weight_decay=self.config.train.weight_decay,
            )
            total = self.total_steps
            n_steps = int(total * self.config.train.warmup)
            self._scheduler = get_linear_schedule_with_warmup(
                self._optim, num_warmup_steps=n_steps, num_training_steps=total
            )

    @property
    def forward_context(self):
        if self.scaler is None:
            return .dummy_context
        else:
            return torch.cuda.amp.autocast

    @property
    def batch_size(self):
        if getattr(self, "_bz", None) is not None:
            return self._bz
        else:
            if self.currently_training():
                return self.config.train.batch_size
            else:
                return self.config.evaluate.batch_size

    # init methods

    def _init_dirs(self):
        if self.config.logdir is not None and self.config.logging:
            os.makedirs(self.config.logdir, exist_ok=True)

    def _init_seed(self):
        random.seed(self.config.seed)
        torch.manual_seed(self.config.seed)

    def _save_outputs(self, save_outputs):
        save_name = os.path.join(
            self.config.logdir, f"user_saved_epoch_{self.cur_epoch}.json"
        )
        json.dump(save_outputs, open(save_name, "w"))

    # def _init_datasets(self):
    #     # first check to see if any train or any val
    #     # check for train and eval loops
    #     self.label_to_id = {}
    #     self.dataset2splits = defaultdict(set)
    #     self.train_visnlangdatasetadapterdict = defaultdict(dict)
    #     self.eval_visnlangdatasetadapterdict = defaultdict(dict)
    #     self.train_visndatasetadapterdict = defaultdict(dict)
    #     self.eval_visndatasetadapterdict = defaultdict(dict)
    #     if not self.config.data.annotations:
    #         self.annotationdict = None
    #     else:
    #         self.annotationdict = {}

    #     train_datasets = set()
    #     eval_datasets = set()
    #     train_splits = set()
    #     eval_splits = set()

    #     # loop through train datasets
    #     for (dset, splits) in self.datasets:
    #         train_datasets.add(dset)
    #         train_splits = train_splits.union(splits)
    #         self.dataset2splits[dset] = self.dataset2splits[dset].union(splits)
    #     # then loop thorugh eval datasets
    #     for (dset, splits) in self.config.data.eval_datasets:
    #         assert (
    #             dset in train_datasets
    #         ), "eval datasets must also be present in train datasets"
    #         eval_datasets.add(dset)
    #         eval_splits = eval_splits.union(splits)
    #         if not splits.intersection(self.dataset2splits[dset]):
    #             self.dataset2splits[dset] = self.dataset2splits[dset].union(splits)

    #     label_id = 0
    #     for name in sorted(set(self.dataset2splits.keys())):
    #         for split in sorted(set(self.dataset2splits[name])):
    #             if (
    #                 name in eval_datasets
    #                 and split in eval_splits
    #                 and not self.config.data.skip_eval
    #             ) or (split in train_splits and not self.config.data.skip_train):
    #                 visnlangdatasetadapter = _visnlangdatasetadapters.get(name).from_config(
    #                     self.config.data, splits=split
    #                 )[split]
    #             else:
    #                 continue
    #             for l in sorted(visnlangdatasetadapter.labels):
    #                 if l not in self.label_to_id:
    #                     self.label_to_id[l] = label_id
    #                     label_id += 1
    #             print(f"Added VisnLangDatasetAdapter {name}: {split}")
    #             if (
    #                 name in eval_datasets
    #                 and split in eval_splits
    #                 and not self.config.data.skip_eval
    #             ):
    #                 self.eval_visnlangdatasetadapterdict[name][split] = visnlangdatasetadapter
    #             if split in train_splits and not self.config.data.skip_train:
    #                 self.train_visnlangdatasetadapterdict[name][split] = visnlangdatasetadapter
    #             is_name, is_split = zip(*visnlangdatasetadapter.data_info[split].items())
    #             is_name = is_name[0]
    #             is_split = is_split[0][0]
    #             if self.config.data.extractor is not None:
    #                 is_path = visnlangdatasetadapter.get_arrow_split(
    #                     self.config.data.datadirs, is_split, self.config.data.extractor
    #                 )
    #                 visndatasetadapter = _visndatasetadapters.get(self.config.data.extractor).from_file(
    #                     is_path
    #                 )
    #             else:
    #                 visndatasetadapter = visnlangdatasetadapter.get_imgid_to_raw_path(
    #                     self.config.data.datadirs, is_split
    #                 )
    #             if self.config.data.annotations and is_name not in self.annotationdict:
    #                 # TODO: must fix this later
    #                 path = os.path.join(
    #                     searchdirs[-1], is_name, "annotations/annotations.arrow"
    #                 )
    #                 ### THIS IS WEHRE I COME BACK TOO ###
    #                 self.eval_visndatasetadapterdict[is_name][is_split] = visndatasetadapter

    #             print(f"Added VisnDatasetAdapter {is_name}: {is_split}")

    #             if (
    #                 name in eval_datasets
    #                 and split in eval_splits
    #                 and not self.config.data.skip_eval
    #             ):
    #                 self.eval_visndatasetadapterdict[is_name][is_split] = visndatasetadapter
    #             if split in train_splits and not self.config.data.skip_train:
    #                 self.train_visndatasetadapterdict[is_name][is_split] = visndatasetadapter

    #     label_file = self.config.data.labels
    #     if label_file is not None or "":
    #         self.label_to_id = json.load(open(label_file))

    # vanilla mehtods
    def write_epoch(self, info: dict = None):
        logstr = ""
        for k, v in info.items():
            logstr += f"{k}={v}; "
        if self.config.logging and info is not None and info:
            logfile = os.path.join(self.config.logdir, "epoch_log.txt")
            if self.currently_training():
                desc = "train"
            else:
                desc = "eval"
            if os.path.isfile(logfile):
                open_type = "a"
            else:
                open_type = "w"
            with open(logfile, open_type) as f:
                date = datetime.datetime.now()
                out_str = f"{desc} | epoch: {self.cur_epoch} | date: {date} | {info} \n"
                f.write(out_str)
                f.flush()
            return True
        return False

    def write_iter(self, info: dict = None, inner_step=0):
        logstr = ""
        for k, v in info.items():
            logstr += f"{k}={v}; "
        if self.config.logging and info is not None and info:
            logfile = os.path.join(self.config.logdir, "steps_log.json")
            assert logfile is not None
            if self.currently_training():
                desc = "train"
            else:
                desc = "eval"
            open_type = "a" if self.cur_step > 1 else "w"
            with open(logfile, open_type) as f:

                date = datetime.datetime.now()
                json.dump(
                    {
                        "desc": desc,
                        "step": self.cur_step,
                        "data": info,
                        "time": str(date),
                    },
                    f,
                )
                f.flush()
            return True
        return False

    def save(self):
        print("\nsaving...\n")
        if self.model_dict:
            for name, model in self.model_dict.items():
                save_name = name + f"_epoch_{self.cur_epoch}.pt"
                save_name = os.path.join(self.config.logdir, save_name)
                torch.save(model.state_dict(), save_name)
        save_name = f"optim_epoch_{self.cur_epoch}.pt"
        if getattr(self, "optim", None) is not None:
            save_name = os.path.join(self.config.logdir, save_name)
            torch.save(self.optim.state_dict(), save_name)
        save_name = os.path.join(self.config.logdir, "info.json")
        json.dump(self.get_exp_info(), open(save_name, "w"))
        save_name = os.path.join(self.config.logdir, "config.yaml")
        self.config.dump_yaml(save_name)
        if hasattr(self, "label_to_id"):
            json.dump(
                self.label_to_id,
                open(os.path.join(self.config.logdir, "labels.json"), "w"),
            )

    def get_exp_info(self):
        exp_info = {
            "name": type(self).__name__,
            "datasets": self.datasets,
            "cur_steps": self.cur_step,
            "cur_epoch": self.cur_epoch,
            "epochs": self.config.train.epochs,
        }

        if getattr(self, "scheduler", None) is not None:
            exp_info["scheduler"]: self.scheduler.state_dict()

        return exp_info

    # dunder methods

    def __call__(self):
        self.outer_loop()

    # helper medhods

    def toTrain(self):
        for k, v in self.__dict__.items():
            if isinstance(v, torch.nn.Module):
                v.train()

    def toEval(self):
        for k, v in self.__dict__.items():
            if isinstance(v, torch.nn.Module):
                v.eval()

    # loop methods

    def outer_loop(self, epoch=None):
        for _ in range(self.epochs):
            if self.cur_epoch == self.epochs + 1:
                return
            # iterate through train and eval loops
            for (run, loader) in self.loaders:
                if loader is None:
                    continue
                else:
                    loop_output = self.inner_loop(
                        loader, train=run, epoch=self.cur_epoch
                    )
                    self.write_epoch(self._clean_dict(loop_output))
            # collect epoch output from each  loop
            if self.config.test_save or self.config.save_after_epoch:
                self.save()

            self.cur_epoch += 1

        if (
            not self.config.save_after_epoch
            and not self.config.test_run
            and self.config.save_after_exp
        ):
            self.save()

    def inner_loop(self, loader, train="train", epoch=None):
        self.__currently_training = train
        desc = "train" if self.currently_training() else "eval"
        # account for cache batch
        if (
            self.config.test_run
            and getattr(loader.dataset, "cache_batch_exists", False)
            and not self.config.data.overwrite_cache_batch
            and self.config.break_loop_on_test
        ):
            loader = [torch.load(loader.dataset.cache_batch_path)]

        _tqdm = tqdm(
            loader,
            desc=f"{desc}_{self.cur_epoch}/{self.config.train.epochs}",
            ncols=0,
            file=sys.stdout,
        )
        loop_outputs = defaultdict(list)
        save_outputs = defaultdict(list)
        if self.currently_training():
            self.toTrain()
        else:
            self.toEval()
        with torch.no_grad() if not self.currently_training() else .dummy_context():
            for inner_step, batch in enumerate(_tqdm):
                if self.currently_training() and self.model_dict:
                    self.optim.zero_grad()
                self.set_batch_size(batch)

                with self.forward_context():
                    forward_outputs = self.forward(batch)
                    assert isinstance(
                        forward_outputs, dict
                    ), "forward method must return dict"

                outputs = self._clean_dict(
                    self.iter_tqdm(forward_outputs, train=self.currently_training())
                )

                # something is wrong with this
                # save_outputs = self._clean_dict(
                #     self.iter_save(forward_outputs, train=train)
                # )
                temp_save_outputs = outputs
                _tqdm.set_postfix(epch=self.cur_epoch, **outputs)

                self.write_iter(outputs, inner_step)

                if self.currently_training():
                    losses = forward_outputs.get("losses", None)
                    if losses is None:
                        pass
                    else:
                        self.step(losses)
                self.cur_step += 1
                # handle loop outputs
                if outputs is not None and loop_outputs is not None:
                    for k, v in outputs.items():
                        loop_outputs[k].append(v)
                else:
                    loop_outputs = None
                # handle saveoutputs
                if temp_save_outputs is not None:
                    for k, v in temp_save_outputs.items():
                        save_outputs[k].append(v)
                # free empty space
                self.garbage_collect()
                if (
                    not self.config.empty_cache
                    and inner_step == 0
                    and self.cur_epoch == 0
                ):
                    pass
                    # torch.cuda.empty_cache()
                # break, possibly if test
                if self.config.test_run and self.config.break_loop_on_test:
                    break

            if save_outputs:
                self._save_outputs(save_outputs)
            # try to take the mean of what we can from the loopoutputs
            for k, v in loop_outputs.items():
                try:
                    loop_outputs[k] = mean(map(lambda x: float(x), v))
                except Exception:
                    pass
            return loop_outputs

    def step(self, loss=None):
        if loss is not None and self.model_dict:
            if self.scaler is not None:
                self.scaler.scale(loss).backward()
                self.scaler.unscale_(self.optim)
                torch.nn..clip_grad_norm_(
                    self.get_grad_params(), self.config.train.max_norm
                )
                self.scaler.step(self.optim)
                self.scaler.update()
            else:
                loss.backward()
                torch.nn..clip_grad_norm_(
                    self.get_grad_params(), self.config.train.max_norm
                )
                self.optim.step()

            self.scheduler.step()

    def get_lr(self):
        if not self.is_train:
            return []
        elif self.scheduler is not None:
            return self.scheduler.get_lr()
        else:
            lrs = []
            for param_group in self.optim.param_groups:
                lrs.append(param_group["lr"])
            return lrs

    @property
    def model_configs(self):
        return self._model_dict

    @property
    def model_dict(self):
        return self._model_dict

    @property
    def loaders(self):
        return self._loaders

    @property
    def scheduler(self):
        return getattr(self, "_scheduler", None)

    @property
    def optim(self):
        return getattr(self, "_optim", None)

    @property
    def total_steps(self):
        if self.is_train:
            total_steps = 0
            for (run, l) in self.loaders:
                if run == "train" and l is not None:
                    total_steps += len(l)

            return self.config.train.epochs * total_steps
        else:
            return len(self.loader)

    def set_batch_size(self, batch):
        for k, v in batch.items():
            if isinstance(v, torch.Tensor):
                self._bz = v.size(0)
                break

    def get_grad_params(self):
        parameters = []
        for k, v in self.model_dict.items():
            parameters.extend([p for p in v.parameters() if p.requires_grad])
        return parameters

    def toCuda(batch, device):
        # will retun batch, but is still in place)
        return .change_device(batch, device)

    def init_grad(name, model):
        pass

    # abstract methods

    @abstractmethod
    def forward(self, batch) -> dict:
        """
        here the user defines the forward run of the model
        returns a dictionary of model outputs
        """

    @abstractmethod
    def iter_tqdm(self, iter_outputs, train=True):
        """
        here the user defines how they would like to configure a dictionary
        that will subesequently be processed and displayed on tqdm
        (this will also be saved to a temp log file)
        """

    @abstractmethod
    def iter_save(iter_outputs, train=True):
        """
        here the user defines what they would like to save from each iteraction
        they can conditionally save if train is true of false
        """

    @abstractmethod
    def epoch_logstr(self, loop_outputs, train=True):
        """
        here the user defines how they would like to configure a dictionary
        that will subesequently be processed and written to a log
        """

    @property
    @abstractmethod
    def model_list(self):
        """
        user defines a list of models. each list item is either a string that references
        a model available in the  library, or a tuple containing the model name and then the model class
        """



from collections import Counter

import datasets as ds
from tqdm import tqdm
from vltk.abc.visnlangdatasetadapter import VisnLangDatasetAdapter

# user must only define forward in this function, and dataset features


class GQAset(VisnLangDatasetAdapter):
    name = "gqa"
    data_info = {
        "dev": {"coco2014": ["test"]},
        "train": {"vg": ["train"]},
        "val": {"vg": ["train"]},
        "test": {"coco2014": ["test"]},
    }
    default_features = {
        "qid": ds.Value("string"),
        "structure": ds.Value("string"),
        "super_class": ds.Value("string"),
        "operations": ds.Sequence(length=-1, feature=ds.Value("string")),
    }
    filters = ["all"]

    def forward(text_data, split, label_preprocessor=None, **kwargs):
        skipped = 0
        min_label_frequency = kwargs.get("min_label_frequency")
        label_frequencies = Counter()
        batch_entries = []
        if label_preprocessor is None:

            def label_preprocessor(x):
                return x

        for t in text_data:
            for i, (k, v) in tqdm(enumerate(t.items())):
                if "answer" in v:
                    answer = label_preprocessor(v["answer"])
                    label_frequencies.update([answer])

            for i, (k, v) in enumerate(t.items()):
                if split == "test":
                    answer = None
                    operations = [""]
                    super_class = ""
                    structure = ""
                elif label_frequencies[v["answer"]] < min_label_frequency:
                    skipped += 1
                    continue
                else:
                    answer = label_preprocessor(v["answer"])
                    operations = [sv["operation"] for sv in v["semantic"]]
                    super_class = v["groups"]["global"]
                    structure = v["types"]["structural"]
                    if super_class is None:
                        super_class = ""

                text = v["question"]
                img_id = v["imageId"].lstrip("n")
                qid = k
                entry = {
                    "qid": qid,
                    "structure": structure,
                    "super_class": super_class,
                    "operations": operations,
                    VisnLangDatasetAdapter.text_key: text,
                    VisnLangDatasetAdapter.img_key: img_id,
                    VisnLangDatasetAdapter.label_key: [answer],
                    VisnLangDatasetAdapter.score_key: [1.0],
                }

                batch_entries.append(entry)

        print(f"SKIPPEd {skipped} entries")
        return batch_entries


if __name__ == "__main__":

    from vltk.configs import Config

    config = Config().data
    GQAset.extract(
        config=config,
        splits=["dev", "test", "val", "train"],
    )
    config.min_label_frequency = 1
    train = GQAset.from_config(config, splits="train")["train"]
    # get min frequency of answers when loading, so we know the lenngth right away

    # get path to arrow file
    # val.get_arrow_split(datadirs=config.datadirs, extractor="frcnn", split="val")
    # get path to raw img ids
    print(len(train.labels))

    # print("entry at row 1:", val.get_row(1))
    # print("entries with img id 262148:", val.get_from_img("262148"))
    # print("freq of answer table:", val.get_freq("table"))
    # print("num_labels", val.num_labels)
